from __future__ import annotations

import gzip
from pathlib import Path

import pytest

from worai.core import sitemap


class _Resp:
    def __init__(self, content: bytes, headers: dict[str, str] | None = None) -> None:
        self.content = content
        self.headers = headers or {}


class _Session:
    def __init__(self, response: _Resp | Exception) -> None:
        self._response = response

    def get(self, _url: str, timeout: float):
        assert timeout == 1.0
        if isinstance(self._response, Exception):
            raise self._response
        return self._response


def _urlset(*urls: str) -> bytes:
    body = "".join(f"<url><loc>{u}</loc></url>" for u in urls)
    return f"<urlset>{body}</urlset>".encode("utf-8")


def _index(*sitemaps: str) -> bytes:
    body = "".join(f"<sitemap><loc>{u}</loc></sitemap>" for u in sitemaps)
    return f"<sitemapindex>{body}</sitemapindex>".encode("utf-8")


def test_helpers_strip_decode_and_parse() -> None:
    assert sitemap._strip_ns("{x}urlset") == "urlset"
    assert sitemap._strip_ns("urlset") == "urlset"

    gz = gzip.compress(_urlset("https://example.com/a"))
    assert sitemap._decode_content("https://e.com/sitemap.xml.gz", _Resp(gz))
    assert sitemap._decode_content("https://e.com/sitemap.xml", _Resp(gz, {"content-type": "application/gzip"}))

    root_tag, locs = sitemap._parse_sitemap_xml(_urlset("https://example.com/a"))
    assert root_tag == "urlset"
    assert locs == ["https://example.com/a"]


def test_parse_sitemap_urls_from_local_index_and_file_scheme(tmp_path: Path) -> None:
    nested = tmp_path / "nested.xml"
    nested.write_bytes(_urlset("https://example.com/one", "https://example.com/two"))

    root = tmp_path / "root.xml"
    root.write_bytes(_index(str(nested)))

    assert sitemap.parse_sitemap_urls(str(root), session=_Session(_Resp(b"")), timeout=1.0) == [
        "https://example.com/one",
        "https://example.com/two",
    ]

    gz_file = tmp_path / "g.xml.gz"
    gz_file.write_bytes(gzip.compress(_urlset("https://example.com/gz")))
    assert sitemap.parse_sitemap_urls(f"file://{gz_file}", session=_Session(_Resp(b"")), timeout=1.0) == [
        "https://example.com/gz"
    ]


def test_parse_sitemap_urls_requests_browser_and_auto(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(sitemap, "_fetch_sitemap_requests", lambda *_a, **_k: _urlset("https://example.com/r"))
    urls = sitemap.parse_sitemap_urls("https://example.com/sitemap.xml", session=_Session(_Resp(b"")), timeout=1.0)
    assert urls == ["https://example.com/r"]

    monkeypatch.setattr(sitemap, "_fetch_sitemap_browser", lambda *_a, **_k: _urlset("https://example.com/b"))
    urls = sitemap.parse_sitemap_urls(
        "https://example.com/sitemap.xml",
        session=_Session(_Resp(b"")),
        timeout=1.0,
        fetch_mode="browser",
    )
    assert urls == ["https://example.com/b"]

    def _raise(*_a, **_k):
        raise RuntimeError("requests failed")

    monkeypatch.setattr(sitemap, "_fetch_sitemap_requests", _raise)
    urls = sitemap.parse_sitemap_urls(
        "https://example.com/sitemap.xml",
        session=_Session(_Resp(b"")),
        timeout=1.0,
        fetch_mode="auto",
    )
    assert urls == ["https://example.com/b"]


def test_parse_sitemap_urls_errors_and_limits(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    with pytest.raises(ValueError, match="Local sitemap file not found"):
        sitemap.parse_sitemap_urls("missing.xml", session=_Session(_Resp(b"")), timeout=1.0)

    with pytest.raises(ValueError, match="Unsupported sitemap URL scheme"):
        sitemap.parse_sitemap_urls("ftp://example.com/x.xml", session=_Session(_Resp(b"")), timeout=1.0)

    monkeypatch.setattr(sitemap, "_fetch_sitemap_requests", lambda *_a, **_k: b"<feed></feed>")
    with pytest.raises(ValueError, match="Unsupported sitemap type"):
        sitemap.parse_sitemap_urls("https://example.com/sitemap.xml", session=_Session(_Resp(b"")), timeout=1.0)

    def _raise(*_a, **_k):
        raise RuntimeError("r")

    def _raise_browser(*_a, **_k):
        raise RuntimeError("b")

    monkeypatch.setattr(sitemap, "_fetch_sitemap_requests", _raise)
    monkeypatch.setattr(sitemap, "_fetch_sitemap_browser", _raise_browser)
    with pytest.raises(RuntimeError, match="Requests sitemap fetch failed"):
        sitemap.parse_sitemap_urls(
            "https://example.com/sitemap.xml",
            session=_Session(_Resp(b"")),
            timeout=1.0,
            fetch_mode="auto",
        )

    monkeypatch.setattr(
        sitemap,
        "_fetch_sitemap_requests",
        lambda *_a, **_k: _urlset("https://example.com/1", "https://example.com/2"),
    )
    urls = sitemap.parse_sitemap_urls(
        "https://example.com/sitemap.xml",
        session=_Session(_Resp(b"")),
        timeout=1.0,
        max_urls=1,
    )
    assert urls == ["https://example.com/1"]

    assert sitemap.get_base_url("https://example.com/path") == "https://example.com"



from __future__ import annotations

from types import SimpleNamespace

import pytest

from worai import cli as mod
from worai.errors import WoraError


def test_truthy_and_should_check_updates(monkeypatch: pytest.MonkeyPatch) -> None:
    assert mod._is_truthy("true")
    assert mod._is_truthy("1")
    assert not mod._is_truthy("0")
    assert not mod._is_truthy(None)

    ctx = SimpleNamespace(obj={"quiet": False})
    monkeypatch.delenv("PYTEST_CURRENT_TEST", raising=False)
    monkeypatch.delenv("WORAI_DISABLE_UPDATE_CHECK", raising=False)
    monkeypatch.setattr(mod.sys, "argv", ["worai", "graph"])
    assert mod._should_check_updates(ctx)

    monkeypatch.setenv("WORAI_DISABLE_UPDATE_CHECK", "1")
    assert not mod._should_check_updates(ctx)


def test_emit_update_notice_and_run_error(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(mod, "_should_check_updates", lambda _ctx: True)
    monkeypatch.setattr(mod, "check_for_update", lambda *_a, **_k: SimpleNamespace(update_available=True, latest_version="9.9.9", current_version="1.0.0", upgrade_command=["pip", "install", "-U", "worai"]))
    seen = {}
    monkeypatch.setattr(mod.typer, "secho", lambda msg, **_k: seen.setdefault("msg", msg))
    mod._emit_update_notice(SimpleNamespace(obj={}))
    assert "Update available for worai" in seen["msg"]

    monkeypatch.setattr(mod, "_should_check_updates", lambda _ctx: False)
    mod._emit_update_notice(SimpleNamespace(obj={}))

    monkeypatch.setattr(mod, "app", lambda: (_ for _ in ()).throw(WoraError("boom", exit_code=7)))
    with pytest.raises(SystemExit) as exc:
        mod.run()
    assert exc.value.code == 7



from __future__ import annotations

from types import SimpleNamespace

import requests

from worai.seocheck.checks.canonical import CanonicalCheck
from worai.seocheck.checks.llms_txt import check_llms_txt
from worai.seocheck.checks.page_meta import PageMetaCheck
from worai.seocheck.checks.resource_404 import Resource404Check
from worai.seocheck.checks.response_time import ResponseTimeCheck
from worai.seocheck.checks.robots_txt import _disallow_all, _extract_groups, check_robots_txt
from worai.seocheck.checks.status import StatusCheck
from worai.seocheck.checks.ttfb import TtfbCheck
from worai.seocheck.checks.wordlift_bootstrap import WordLiftBootstrapCheck


class _Page:
    def __init__(self, *, title: str = "", values: dict[str, object] | None = None, raises: bool = False) -> None:
        self._title = title
        self._values = values or {}
        self._raises = raises

    def title(self) -> str:
        return self._title

    def evaluate(self, script: str, *_args):
        if self._raises:
            raise RuntimeError("eval failed")
        for key, value in self._values.items():
            if key in script:
                return value
        return self._values.get("default", "")


class _Session:
    def __init__(self, responses):
        self._responses = list(responses)

    def get(self, _url: str, timeout: float):
        assert timeout == 1.0
        if not self._responses:
            raise AssertionError("no response")
        response = self._responses.pop(0)
        if isinstance(response, Exception):
            raise response
        return response


def test_canonical_status_response_time_and_resource_checks() -> None:
    canonical_ok = CanonicalCheck().run(
        page=_Page(values={"canonical": "https://example.com/"}),
        response=None,
        elapsed_ms=0,
        resources=[],
    )
    assert canonical_ok.status == "ok"

    canonical_warn = CanonicalCheck().run(page=_Page(), response=None, elapsed_ms=0, resources=[])
    assert canonical_warn.status == "warn"

    assert ResponseTimeCheck().run(page=None, response=None, elapsed_ms=100, resources=[]).status == "ok"
    assert ResponseTimeCheck().run(page=None, response=None, elapsed_ms=2500, resources=[]).status == "warn"
    assert ResponseTimeCheck().run(page=None, response=None, elapsed_ms=7000, resources=[]).status == "fail"

    resources = [
        {"status": 200, "resource_type": "image"},
        {"status": 404, "resource_type": "image", "url": "x"},
        {"status": 500, "resource_type": "script", "url": "y"},
        {"status": 403, "resource_type": "xhr", "url": "z"},
    ]
    result = Resource404Check(max_samples=1).run(page=None, response=None, elapsed_ms=0, resources=resources)
    assert result.status == "warn"
    assert result.data["count"] == 2
    assert len(result.data["samples"]) == 1


def test_page_meta_status_ttfb_and_wordlift_checks() -> None:
    page = _Page(
        title="Title",
        values={
            "meta[name=\"description\"]": "Desc",
            "meta[name=\"robots\"]": "index,follow",
            "meta[property=\"og:title\"]": "OGT",
            "meta[property=\"og:description\"]": "OGD",
        },
    )
    meta_ok = PageMetaCheck().run(page=page, response=None, elapsed_ms=0, resources=[])
    assert meta_ok.status == "ok"

    meta_warn = PageMetaCheck().run(page=_Page(), response=None, elapsed_ms=0, resources=[])
    assert meta_warn.status == "warn"
    assert "missing title" in meta_warn.details

    assert StatusCheck().run(page=None, response=None, elapsed_ms=0, resources=[]).status == "fail"
    assert StatusCheck().run(page=None, response=SimpleNamespace(status=200), elapsed_ms=0, resources=[]).status == "ok"
    assert StatusCheck().run(page=None, response=SimpleNamespace(status=500), elapsed_ms=0, resources=[]).status == "warn"

    ttfb = TtfbCheck(ok_ms=100, warn_ms=200)
    assert ttfb.run(page=_Page(), response=None, elapsed_ms=0, resources=[]).status == "fail"
    assert ttfb.run(page=_Page(raises=True), response=SimpleNamespace(status=200), elapsed_ms=0, resources=[]).status == "fail"
    assert ttfb.run(page=_Page(values={"default": None}), response=SimpleNamespace(status=200), elapsed_ms=0, resources=[]).status == "fail"
    assert (
        ttfb.run(
            page=_Page(values={"default": {"requestStart": 1, "responseStart": 50}}),
            response=SimpleNamespace(status=200),
            elapsed_ms=0,
            resources=[],
        ).status
        == "ok"
    )
    assert (
        ttfb.run(
            page=_Page(values={"default": {"requestStart": 1, "responseStart": 250}}),
            response=SimpleNamespace(status=200),
            elapsed_ms=0,
            resources=[],
        ).status
        == "fail"
    )

    assert WordLiftBootstrapCheck().run(
        page=_Page(values={"default": True}),
        response=None,
        elapsed_ms=0,
        resources=[],
    ).status == "ok"
    assert WordLiftBootstrapCheck().run(
        page=_Page(values={"default": False}),
        response=None,
        elapsed_ms=0,
        resources=[],
    ).status == "fail"


def test_llms_and_robots_checks_and_parsers() -> None:
    assert _disallow_all("User-agent: *\nDisallow: /\n") is True
    assert _disallow_all("User-agent: *\nDisallow:\n") is False

    groups = _extract_groups(["User-agent: *", "Disallow: /", "User-agent: test", "Disallow: /no"]) 
    assert len(groups) == 2

    fail = check_llms_txt("https://example.com", session=_Session([requests.RequestException("x")]), timeout=1.0)
    assert fail.status == "fail"

    ok = check_llms_txt(
        "https://example.com",
        session=_Session([SimpleNamespace(status_code=200, text="abc")]),
        timeout=1.0,
    )
    assert ok.status == "ok"

    warn = check_llms_txt(
        "https://example.com",
        session=_Session([SimpleNamespace(status_code=404, text=""), SimpleNamespace(status_code=404, text="")]),
        timeout=1.0,
    )
    assert warn.status == "warn"

    robots_fail = check_robots_txt(
        "https://example.com",
        session=_Session([requests.RequestException("x")]),
        timeout=1.0,
    )
    assert robots_fail.status == "fail"

    robots_warn = check_robots_txt(
        "https://example.com",
        session=_Session([SimpleNamespace(status_code=500, text="")]),
        timeout=1.0,
    )
    assert robots_warn.status == "warn"

    robots_ok = check_robots_txt(
        "https://example.com",
        session=_Session([SimpleNamespace(status_code=200, text="User-agent: *\nDisallow:" )]),
        timeout=1.0,
    )
    assert robots_ok.status == "ok"

    robots_blocked = check_robots_txt(
        "https://example.com",
        session=_Session([SimpleNamespace(status_code=200, text="User-agent: *\nDisallow: /")]),
        timeout=1.0,
    )
    assert robots_blocked.status == "warn"



from __future__ import annotations

from dataclasses import dataclass
import logging
from pathlib import Path
import re
import subprocess

from typer.testing import CliRunner

from worai import cli as root_cli
from worai.commands import graph as graph_cmd
from worai.core.graph import create as graph_create
from worai.core.graph import sync as graph_sync
from worai.errors import UsageError


@dataclass
class _FakeProfile:
    name: str
    api_key: str | None
    settings: dict


_ANSI_ESCAPE_RE = re.compile(r"\x1b\[[0-9;]*m")


def _normalize_cli_output(text: str) -> str:
    return _ANSI_ESCAPE_RE.sub("", text)


def _patch_sdk_availability(monkeypatch) -> None:
    class _ConfigurationProvider:
        @staticmethod
        def create(_path=None):
            return None

    monkeypatch.setattr(graph_sync, "ConfigurationProvider", _ConfigurationProvider)
    monkeypatch.setattr(graph_sync, "KgBuildApplicationContainer", object())
    monkeypatch.setattr(graph_sync, "ProfileImportProtocol", object())


def test_graph_sync_help_from_root_cli() -> None:
    runner = CliRunner()
    result = runner.invoke(root_cli.app, ["graph", "sync", "run", "--help"])
    output = _normalize_cli_output(result.output)

    assert result.exit_code == 0
    assert "--profile" in output
    assert "--debug" in output


def test_graph_sync_requires_profile_option() -> None:
    runner = CliRunner()
    result = runner.invoke(graph_cmd.app, ["sync", "run"])
    output = _normalize_cli_output(result.output)

    assert result.exit_code == 2
    assert "--profile" in output


def test_graph_sync_command_calls_core(monkeypatch, tmp_path: Path) -> None:
    seen = {}

    def _stub_run(options):
        seen["options"] = options

    monkeypatch.setattr(graph_cmd, "run_graph_sync", _stub_run)
    monkeypatch.setattr(graph_cmd, "resolve_config_path", lambda _path: Path("worai.toml"))
    runner = CliRunner()
    result = runner.invoke(graph_cmd.app, ["sync", "run", "--profile", "zrh", "--debug"])

    assert result.exit_code == 0
    assert seen["options"].profile == "zrh"
    assert seen["options"].config_path == Path("worai.toml")
    assert seen["options"].debug is True


def test_graph_sync_uses_root_config_option(monkeypatch, tmp_path: Path) -> None:
    seen = {}

    def _stub_run(options):
        seen["options"] = options

    monkeypatch.setattr(graph_cmd, "run_graph_sync", _stub_run)
    runner = CliRunner()
    cfg = tmp_path / "custom.toml"
    cfg.write_text("[profile.zrh]\napi_key='wl'\n", encoding="utf-8")
    result = runner.invoke(
        root_cli.app,
        ["--config", str(cfg), "graph", "sync", "run", "--profile", "zrh"],
    )

    assert result.exit_code == 0
    assert seen["options"].config_path == cfg


def test_graph_sync_help_lists_subcommands() -> None:
    runner = CliRunner()
    result = runner.invoke(root_cli.app, ["graph", "sync", "--help"])
    output = _normalize_cli_output(result.output)

    assert result.exit_code == 0
    assert "run" in output
    assert "create" in output


def test_graph_sync_create_command_calls_core(monkeypatch) -> None:
    seen = {}

    def _stub_create(options):
        seen["options"] = options

    monkeypatch.setattr(graph_cmd, "run_graph_sync_create", _stub_create)
    runner = CliRunner()
    result = runner.invoke(
        graph_cmd.app,
        [
            "sync",
            "create",
            "./demo-project",
            "--template",
            "./template",
            "--defaults",
            "--non-interactive",
            "--force",
        ],
    )

    assert result.exit_code == 0
    assert seen["options"].destination == Path("./demo-project")
    assert seen["options"].template == "./template"
    assert seen["options"].defaults is True
    assert seen["options"].non_interactive is True
    assert seen["options"].force is True


def test_graph_sync_create_requires_destination_in_non_interactive() -> None:
    runner = CliRunner()
    result = runner.invoke(graph_cmd.app, ["sync", "create", "--non-interactive"])

    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)
    assert "Destination path is required" in str(result.exception)


def test_load_service_account_json_supports_file_or_inline(tmp_path: Path) -> None:
    service_account = '{"type":"service_account"}'
    path = tmp_path / "service-account.json"
    path.write_text(service_account, encoding="utf-8")

    assert graph_sync.load_service_account_json(str(path)) == service_account
    assert graph_sync.load_service_account_json(service_account) == service_account


def test_load_service_account_json_rejects_invalid_inputs() -> None:
    try:
        graph_sync.load_service_account_json("not-json")
        assert False, "Expected invalid JSON/path error"
    except ValueError as exc:
        assert "json content or a file path" in str(exc).lower()

    try:
        graph_sync.load_service_account_json("[]")
        assert False, "Expected non-object JSON error"
    except ValueError as exc:
        assert "must be an object" in str(exc).lower()


def test_parse_bool_and_timeout_guardrails() -> None:
    assert graph_sync._parse_bool("yes", name="x") is True
    assert graph_sync._parse_bool("off", name="x") is False
    try:
        graph_sync._parse_bool("maybe", name="x")
        assert False, "Expected invalid bool"
    except ValueError:
        pass

    for bad in ["", object(), 0]:
        try:
            graph_sync._resolve_web_page_import_timeout_ms(bad)
            assert False, "Expected invalid timeout"
        except ValueError:
            pass


def test_build_dynamic_settings_supports_sitemap_source() -> None:
    lines, sheets_json = graph_sync._build_dynamic_settings(
        "wl_key",
        {
            "sitemap_url": "https://example.com/sitemap.xml",
            "sitemap_url_pattern": "example\\.com",
            "concurrency": 3,
        },
        google_search_console=False,
    )

    text = "\n".join(lines)
    assert "SITEMAP_URL" in text
    assert "SITEMAP_URL_PATTERN" in text
    assert "INGEST_SOURCE = 'sitemap'" in text
    assert "INGEST_LOADER = 'web_scrape_api'" in text
    assert "INGEST_TIMEOUT_MS = 60000" in text
    assert sheets_json is None


def test_build_dynamic_settings_does_not_require_sheets_account_for_urls() -> None:
    lines, sheets_json = graph_sync._build_dynamic_settings(
        "wl_key",
        {
            "urls": ["https://example.com/a"],
            "sheets_service_account": "not-json-and-not-file",
        },
        google_search_console=False,
    )

    text = "\n".join(lines)
    assert "URLS" in text
    assert "INGEST_SOURCE = 'urls'" in text
    assert sheets_json is None


def test_build_dynamic_settings_prefers_new_ingest_loader_over_legacy() -> None:
    lines, _ = graph_sync._build_dynamic_settings(
        "wl_key",
        {
            "urls": ["https://example.com/a"],
            "web_page_import_mode": "proxy",
            "ingest_loader": "web_scrape_api",
        },
        google_search_console=False,
    )

    assert "INGEST_LOADER = 'web_scrape_api'" in lines
    assert "INGEST_LOADER = 'web_scrape_api'" in lines


def test_build_dynamic_settings_maps_premium_scraper_loader_to_matching_fetch_mode() -> None:
    lines, _ = graph_sync._build_dynamic_settings(
        "wl_key",
        {
            "urls": ["https://example.com/a"],
            "ingest_loader": "premium_scraper",
        },
        google_search_console=False,
    )

    assert "INGEST_LOADER = 'premium_scraper'" in lines
    assert "INGEST_LOADER = 'premium_scraper'" in lines


def test_build_dynamic_settings_maps_proxy_loader_to_matching_fetch_mode() -> None:
    lines, _ = graph_sync._build_dynamic_settings(
        "wl_key",
        {
            "urls": ["https://example.com/a"],
            "ingest_loader": "proxy",
        },
        google_search_console=False,
    )

    assert "INGEST_LOADER = 'proxy'" in lines
    assert "INGEST_LOADER = 'proxy'" in lines


def test_build_dynamic_settings_legacy_web_page_import_mode_still_supported() -> None:
    lines, _ = graph_sync._build_dynamic_settings(
        "wl_key",
        {
            "urls": ["https://example.com/a"],
            "web_page_import_mode": "premium_scraper",
        },
        google_search_console=False,
    )

    assert "INGEST_LOADER = 'premium_scraper'" in lines
    assert "INGEST_LOADER = 'premium_scraper'" in lines


def test_build_dynamic_settings_converts_timeout_seconds_to_ms() -> None:
    lines, _ = graph_sync._build_dynamic_settings(
        "wl_key",
        {
            "urls": ["https://example.com/a"],
            "web_page_import_timeout": 60,
        },
        google_search_console=False,
    )

    assert any(line == "INGEST_TIMEOUT_MS = 60000" for line in lines)


def test_build_dynamic_settings_accepts_timeout_suffixes() -> None:
    lines_seconds, _ = graph_sync._build_dynamic_settings(
        "wl_key",
        {
            "urls": ["https://example.com/a"],
            "web_page_import_timeout": "2.5s",
        },
        google_search_console=False,
    )
    lines_millis, _ = graph_sync._build_dynamic_settings(
        "wl_key",
        {
            "urls": ["https://example.com/a"],
            "web_page_import_timeout": "2500ms",
        },
        google_search_console=False,
    )

    assert any(line == "INGEST_TIMEOUT_MS = 2500" for line in lines_seconds)
    assert any(line == "INGEST_TIMEOUT_MS = 2500" for line in lines_millis)


def test_build_dynamic_settings_treats_plain_numeric_timeout_as_seconds() -> None:
    lines, _ = graph_sync._build_dynamic_settings(
        "wl_key",
        {
            "urls": ["https://example.com/a"],
            "web_page_import_timeout": 60000,
        },
        google_search_console=False,
    )

    assert any(line == "INGEST_TIMEOUT_MS = 60000000" for line in lines)


def test_build_dynamic_settings_rejects_invalid_timeout() -> None:
    try:
        graph_sync._build_dynamic_settings(
            "wl_key",
            {
                "urls": ["https://example.com/a"],
                "web_page_import_timeout": "abc",
            },
            google_search_console=False,
        )
        assert False, "Expected ValueError for invalid timeout format"
    except ValueError as exc:
        assert "web_page_import_timeout" in str(exc)


def test_build_dynamic_settings_requires_sheets_account_for_sheets_source() -> None:
    try:
        graph_sync._build_dynamic_settings(
            "wl_key",
            {
                "sheets_url": "https://docs.google.com/spreadsheets/d/abc",
                "sheets_name": "Sheet1",
            },
            google_search_console=False,
        )
        assert False, "Expected ValueError for missing sheets_service_account"
    except ValueError as exc:
        assert "sheets_service_account" in str(exc)


def test_build_dynamic_settings_requires_both_sheets_fields() -> None:
    try:
        graph_sync._build_dynamic_settings(
            "wl_key",
            {
                "sheets_url": "https://docs.google.com/spreadsheets/d/abc",
                "sheets_service_account": "{\"type\":\"service_account\"}",
            },
            google_search_console=False,
        )
        assert False, "Expected ValueError for missing sheets_name"
    except ValueError as exc:
        assert "both sheets_url and sheets_name" in str(exc).lower()


def test_build_dynamic_settings_auto_source_resolves_to_sheets() -> None:
    lines, _ = graph_sync._build_dynamic_settings(
        "wl_key",
        {
            "ingest_source": "auto",
            "sheets_url": "https://docs.google.com/spreadsheets/d/abc",
            "sheets_name": "Sheet1",
            "sheets_service_account": "{\"type\":\"service_account\"}",
        },
        google_search_console=False,
    )
    assert "INGEST_SOURCE = 'sheets'" in lines


def test_build_dynamic_settings_rejects_multiple_sources() -> None:
    try:
        graph_sync._build_dynamic_settings(
            "wl_key",
            {
                "urls": ["https://example.com/a"],
                "sitemap_url": "https://example.com/sitemap.xml",
            },
            google_search_console=False,
        )
        assert False, "Expected ValueError for multiple source modes"
    except ValueError as exc:
        assert "exactly one source" in str(exc).lower()


def test_resolve_google_search_console_default_false(tmp_path: Path) -> None:
    cfg = tmp_path / "worai.toml"
    cfg.write_text("[profile.acme]\napi_key='wl'\n", encoding="utf-8")

    assert graph_sync.resolve_google_search_console(cfg, "acme") is False


def test_resolve_google_search_console_global_and_profile_override(tmp_path: Path) -> None:
    cfg = tmp_path / "worai.toml"
    cfg.write_text(
        "google_search_console = false\n\n[profile.acme]\napi_key='wl'\ngoogle_search_console = true\n",
        encoding="utf-8",
    )

    assert graph_sync.resolve_google_search_console(cfg, "acme") is True


def test_resolve_google_search_console_supports_profiles_section(tmp_path: Path) -> None:
    cfg = tmp_path / "worai.toml"
    cfg.write_text(
        "google_search_console = false\n\n[profiles.acme]\napi_key='wl'\ngoogle_search_console = true\n",
        encoding="utf-8",
    )

    assert graph_sync.resolve_google_search_console(cfg, "acme") is True


def test_resolve_google_search_console_file_errors(tmp_path: Path) -> None:
    assert graph_sync.resolve_google_search_console(tmp_path / "missing.toml", "acme") is False
    bad = tmp_path / "bad.toml"
    bad.write_text("not = [valid", encoding="utf-8")
    try:
        graph_sync.resolve_google_search_console(bad, "acme")
        assert False, "Expected invalid TOML error"
    except ValueError as exc:
        assert "invalid toml config" in str(exc).lower()


def test_resolve_postprocessor_runtime_default_none(tmp_path: Path) -> None:
    cfg = tmp_path / "worai.toml"
    cfg.write_text("[profile.acme]\napi_key='wl'\n", encoding="utf-8")

    assert graph_sync.resolve_postprocessor_runtime(cfg, "acme") is None


def test_resolve_postprocessor_runtime_global_and_profile_override(tmp_path: Path) -> None:
    cfg = tmp_path / "worai.toml"
    cfg.write_text(
        "postprocessor_runtime = 'oneshot'\n\n"
        "[profile.acme]\napi_key='wl'\npostprocessor_runtime = 'persistent'\n",
        encoding="utf-8",
    )

    assert graph_sync.resolve_postprocessor_runtime(cfg, "acme") == "persistent"


def test_resolve_postprocessor_runtime_supports_profiles_section(tmp_path: Path) -> None:
    cfg = tmp_path / "worai.toml"
    cfg.write_text(
        "postprocessor_runtime = 'oneshot'\n\n"
        "[profiles.acme]\napi_key='wl'\npostprocessor_runtime = 'persistent'\n",
        encoding="utf-8",
    )

    assert graph_sync.resolve_postprocessor_runtime(cfg, "acme") == "persistent"


def test_resolve_postprocessor_runtime_validation_errors(tmp_path: Path) -> None:
    assert graph_sync.resolve_postprocessor_runtime(tmp_path / "missing.toml", "acme") is None
    bad = tmp_path / "bad.toml"
    bad.write_text("not = [valid", encoding="utf-8")
    try:
        graph_sync.resolve_postprocessor_runtime(bad, "acme")
        assert False, "Expected invalid TOML error"
    except ValueError as exc:
        assert "invalid toml config" in str(exc).lower()

    cfg = tmp_path / "invalid-runtime.toml"
    cfg.write_text("[profile.acme]\napi_key='wl'\npostprocessor_runtime = 7\n", encoding="utf-8")
    try:
        graph_sync.resolve_postprocessor_runtime(cfg, "acme")
        assert False, "Expected type validation error"
    except ValueError as exc:
        assert "must be a string" in str(exc).lower()

    cfg.write_text("[profile.acme]\napi_key='wl'\npostprocessor_runtime = 'invalid'\n", encoding="utf-8")
    try:
        graph_sync.resolve_postprocessor_runtime(cfg, "acme")
        assert False, "Expected enum validation error"
    except ValueError as exc:
        assert "oneshot, persistent" in str(exc).lower()


def test_parse_settings_and_validation_formatters() -> None:
    parsed = graph_sync._parse_settings_lines(["A = 1", "ignored-line"])
    assert parsed == {"A": 1}
    assert graph_sync._format_validation_summary(None) == "validation=disabled"
    assert graph_sync._format_validation_summary("x") == "validation=x"
    text = graph_sync._format_validation_summary(
        {
            "total": 2,
            "pass": 1,
            "fail": 1,
            "warnings": {"count": 3},
            "errors": {"count": 4},
        }
    )
    assert "warnings=3" in text
    assert "errors=4" in text


def test_cloud_progress_and_kpi_non_dict_payloads_are_logged(caplog) -> None:
    caplog.set_level(logging.INFO, logger="worai.graph")
    graph_sync._on_cloud_progress("abc")
    graph_sync._on_cloud_kpi("xyz")
    assert "Graph sync progress: abc" in caplog.text
    assert "Graph sync KPI: xyz" in caplog.text


def test_cloud_kpi_invariant_mismatch_warns(caplog) -> None:
    caplog.set_level(logging.WARNING, logger="worai.graph")
    graph_sync._on_cloud_kpi(
        {
            "profile": "zrh",
            "run_id": "r",
            "totals": {},
            "validation": {"total": 3, "pass": 1, "fail": 1},
        }
    )
    assert "KPI validation invariant mismatch" in caplog.text


def test_build_cloud_workflow_config_errors(monkeypatch) -> None:
    class _CloudWorkflowConfig:
        def __init__(self, **_kwargs):
            raise TypeError("boom")

    monkeypatch.setattr(graph_sync, "CloudWorkflowConfig", None)
    try:
        graph_sync._build_cloud_workflow_config(
            lines=["WORDLIFT_KEY = 'wl'", "URLS = ['https://example.com']"],
            sheets_service_account_json=None,
            debug=False,
            debug_profile_name="p",
        )
        assert False, "Expected missing CloudWorkflowConfig error"
    except ValueError as exc:
        assert "cloud workflow config is unavailable" in str(exc).lower()

    monkeypatch.setattr(graph_sync, "CloudWorkflowConfig", _CloudWorkflowConfig)
    try:
        graph_sync._build_cloud_workflow_config(
            lines=["WORDLIFT_KEY = 'wl'"],
            sheets_service_account_json=None,
            debug=False,
            debug_profile_name="p",
        )
        assert False, "Expected source mode error"
    except ValueError as exc:
        assert "exactly one source mode" in str(exc).lower()

    try:
        graph_sync._build_cloud_workflow_config(
            lines=[
                "WORDLIFT_KEY = 'wl'",
                "SHEETS_URL = 'https://docs.google.com/spreadsheets/d/abc'",
            ],
            sheets_service_account_json=None,
            debug=False,
            debug_profile_name="p",
        )
        assert False, "Expected missing sheets_name error"
    except ValueError as exc:
        assert "sheets_url and sheets_name" in str(exc).lower()

    try:
        graph_sync._build_cloud_workflow_config(
            lines=["WORDLIFT_KEY = 'wl'", "URLS = ['https://example.com']"],
            sheets_service_account_json=None,
            debug=False,
            debug_profile_name="p",
        )
        assert False, "Expected sdk contract type error"
    except ValueError as exc:
        assert "expected 6.x cloud workflow contract" in str(exc).lower()


def test_run_cloud_workflow_with_callbacks_support_checks(monkeypatch) -> None:
    async def _protocol_factory(_context, **_kwargs):
        return object()

    monkeypatch.setattr(graph_sync, "ConfigurationProvider", None)
    monkeypatch.setattr(graph_sync, "KgBuildApplicationContainer", None)
    monkeypatch.setattr(graph_sync, "run_cloud_workflow", None)

    try:
        import asyncio

        asyncio.run(
            graph_sync._run_cloud_workflow_with_callbacks(
                lines=["WORDLIFT_KEY = 'wl'", "URLS = ['https://example.com']"],
                sheets_service_account_json=None,
                debug=False,
                debug_profile_name="p",
                protocol_factory=_protocol_factory,
            )
        )
        assert False, "Expected cloud workflow support error"
    except ValueError as exc:
        assert "cloud workflow support is unavailable" in str(exc).lower()


def test_run_graph_sync_async_keeps_cwd_root_and_passes_protocol(monkeypatch) -> None:
    profile = _FakeProfile(
        name="zrh",
        api_key="wl_key",
        settings={
            "overwrite": True,
            "concurrency": 3,
            "web_page_import_mode": "premium_scraper",
            "web_page_import_timeout": 1234,
            "urls": ["https://example.com/a"],
            "custom_toggle": "on",
        },
    )
    captured = {}

    class _ProfileMap(dict):
        def get(self, key):
            return profile if key == "zrh" else None

    class _FakeProtocol:
        def __init__(self, context, profile, root_dir, debug_dir):
            self.context = context
            self.profile = profile
            self.root_dir = root_dir
            self.debug_dir = debug_dir

    class _ConfigurationProvider:
        @staticmethod
        def create(_path=None):
            return None

    class _FakeContainer:
        def __init__(self, _provider):
            self.provider = _provider

    class _CloudWorkflowConfig:
        def __init__(self, **kwargs):
            self.kwargs = kwargs

    async def _stub_run_cloud_workflow(
        *,
        config,
        configuration_provider_create,
        container_factory,
        protocol_factory,
        on_info=None,
        on_progress=None,
        on_kpi=None,
    ):
        captured["config"] = config.kwargs
        captured["postprocessor_runtime"] = graph_sync.os.environ.get("POSTPROCESSOR_RUNTIME")
        _provider = configuration_provider_create("dynamic.py")
        _container = container_factory(_provider)
        protocol = protocol_factory(context={"k": "v"}, debug_dir=Path("dbg"), workflow_config=config)
        if hasattr(protocol, "__await__"):
            protocol = await protocol
        captured["protocol"] = protocol

    monkeypatch.setattr(graph_sync, "ConfigurationProvider", _ConfigurationProvider)
    monkeypatch.setattr(graph_sync, "KgBuildApplicationContainer", _FakeContainer)
    monkeypatch.setattr(graph_sync, "load_profile_config", lambda _path: _ProfileMap())
    monkeypatch.setattr(graph_sync, "ProfileImportProtocol", _FakeProtocol)
    monkeypatch.setattr(graph_sync, "CloudWorkflowConfig", _CloudWorkflowConfig)
    monkeypatch.setattr(graph_sync, "run_cloud_workflow", _stub_run_cloud_workflow)
    monkeypatch.setattr(graph_sync, "resolve_google_search_console", lambda *_args, **_kwargs: True)
    monkeypatch.setattr(graph_sync, "resolve_postprocessor_runtime", lambda *_args, **_kwargs: "persistent")
    monkeypatch.setenv("POSTPROCESSOR_RUNTIME", "oneshot")

    graph_sync.run_graph_sync(
        graph_sync.GraphSyncOptions(
            profile="zrh",
            config_path=Path("worai.toml"),
            debug=True,
        )
    )

    assert captured["config"]["wordlift_key"] == "wl_key"
    assert captured["config"]["urls"] == ["https://example.com/a"]
    assert captured["config"]["ingest_loader"] == "premium_scraper"
    assert captured["config"]["ingest_timeout_ms"] == 1234000
    assert "web_page_import_mode" not in captured["config"]
    assert "web_page_import_timeout" not in captured["config"]
    assert captured["config"]["extra_settings"]["CUSTOM_TOGGLE"] == "on"
    assert captured["config"]["extra_settings"]["GOOGLE_SEARCH_CONSOLE"] is True
    assert captured["postprocessor_runtime"] == "persistent"
    assert captured["config"]["debug"] is True
    assert captured["config"]["debug_profile_name"] == "zrh"
    assert captured["protocol"].root_dir == Path.cwd()
    assert graph_sync.os.environ.get("POSTPROCESSOR_RUNTIME") == "oneshot"


def test_run_graph_sync_async_uses_cloud_flow_callbacks_when_available(monkeypatch, caplog) -> None:
    profile = _FakeProfile(
        name="zrh",
        api_key="wl_key",
        settings={
            "urls": ["https://example.com/a"],
            "overwrite": True,
            "concurrency": 2,
        },
    )
    captured = {}

    class _ProfileMap(dict):
        def get(self, key):
            return profile if key == "zrh" else None

    class _ConfigurationProvider:
        @staticmethod
        def create(path=None):
            captured["provider_path"] = path
            return {"provider_path": path}

    class _FakeContainer:
        def __init__(self, provider):
            self.provider = provider

    class _FakeProtocol:
        def __init__(self, context, profile, root_dir, debug_dir):
            self.context = context
            self.profile = profile
            self.root_dir = root_dir
            self.debug_dir = debug_dir

    class _CloudWorkflowConfig:
        def __init__(self, **kwargs):
            self.kwargs = kwargs

    async def _stub_run_cloud_workflow(
        *,
        config,
        configuration_provider_create,
        container_factory,
        protocol_factory,
        on_info=None,
        on_progress=None,
        on_kpi=None,
    ):
        captured["config"] = config.kwargs
        provider = configuration_provider_create("dynamic.py")
        container = container_factory(provider)
        protocol = protocol_factory(context={"k": "v"}, debug_dir=Path("dbg"), workflow_config=config)
        if hasattr(protocol, "__await__"):
            protocol = await protocol
        captured["container"] = container
        captured["protocol"] = protocol
        if on_info:
            on_info("cloud-info")
        if on_progress:
            on_progress(
                {
                    "kind": "graph",
                    "profile": "zrh",
                    "url": "https://example.com/a",
                    "graph": {"entities": 1, "type_assertions": 2, "property_assertions": 3},
                    "validation": None,
                }
            )
        if on_kpi:
            on_kpi(
                {
                    "run_id": "2026-02-24T15:12:03Z",
                    "profile": "zrh",
                    "totals": {
                        "total_entities": 1,
                        "type_assertions_total": 2,
                        "property_assertions_total": 3,
                    },
                    "entities_by_type": {"https://schema.org/WebPage": 1},
                    "properties_by_predicate": {"https://schema.org/name": 1},
                    "validation": None,
                }
            )

    monkeypatch.setattr(graph_sync, "ConfigurationProvider", _ConfigurationProvider)
    monkeypatch.setattr(graph_sync, "KgBuildApplicationContainer", _FakeContainer)
    monkeypatch.setattr(graph_sync, "ProfileImportProtocol", _FakeProtocol)
    monkeypatch.setattr(graph_sync, "CloudWorkflowConfig", _CloudWorkflowConfig)
    monkeypatch.setattr(graph_sync, "run_cloud_workflow", _stub_run_cloud_workflow)
    monkeypatch.setattr(graph_sync, "load_profile_config", lambda _path: _ProfileMap())
    monkeypatch.setattr(graph_sync, "resolve_google_search_console", lambda *_args, **_kwargs: False)
    monkeypatch.setattr(graph_sync, "resolve_postprocessor_runtime", lambda *_args, **_kwargs: None)

    caplog.set_level(logging.INFO, logger="worai.graph")
    graph_sync.run_graph_sync(
        graph_sync.GraphSyncOptions(
            profile="zrh",
            config_path=Path("worai.toml"),
            debug=True,
        )
    )

    assert captured["config"]["wordlift_key"] == "wl_key"
    assert captured["config"]["urls"] == ["https://example.com/a"]
    assert captured["config"]["debug"] is True
    assert captured["config"]["debug_profile_name"] == "zrh"
    assert captured["protocol"].root_dir == Path.cwd()
    assert "cloud-info" in caplog.text
    assert "Progress[graph] https://example.com/a" in caplog.text
    assert "KPI profile=zrh run_id=2026-02-24T15:12:03Z" in caplog.text


def test_run_graph_sync_async_fails_without_cloud_flow_support(monkeypatch) -> None:
    profile = _FakeProfile(
        name="zrh",
        api_key="wl_key",
        settings={"urls": ["https://example.com/a"]},
    )

    class _ProfileMap(dict):
        def get(self, key):
            return profile if key == "zrh" else None

    monkeypatch.setattr(graph_sync, "ConfigurationProvider", None)
    monkeypatch.setattr(graph_sync, "KgBuildApplicationContainer", None)
    monkeypatch.setattr(graph_sync, "run_cloud_workflow", None)
    monkeypatch.setattr(graph_sync, "load_profile_config", lambda _path: _ProfileMap())
    monkeypatch.setattr(graph_sync, "resolve_google_search_console", lambda *_args, **_kwargs: False)
    monkeypatch.setattr(graph_sync, "resolve_postprocessor_runtime", lambda *_args, **_kwargs: None)

    try:
        graph_sync.run_graph_sync(
            graph_sync.GraphSyncOptions(
                profile="zrh",
                config_path=Path("worai.toml"),
                debug=False,
            )
        )
        assert False, "Expected ValueError when cloud workflow support is unavailable"
    except ValueError as exc:
        assert "unavailable" in str(exc).lower()


def test_run_graph_sync_async_fails_when_profile_has_no_api_key(monkeypatch) -> None:
    profile = _FakeProfile(name="zrh", api_key=None, settings={"urls": ["https://example.com"]})

    class _ProfileMap(dict):
        def get(self, key):
            return profile if key == "zrh" else None

    _patch_sdk_availability(monkeypatch)
    monkeypatch.setattr(graph_sync, "load_profile_config", lambda _path: _ProfileMap())

    runner = CliRunner()
    result = runner.invoke(graph_cmd.app, ["sync", "run", "--profile", "zrh"])

    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)
    assert "missing api_key" in str(result.exception)


def test_graph_sync_create_builds_copier_command_with_unsafe_by_default(tmp_path: Path, monkeypatch) -> None:
    seen = {}

    def _stub_run(cmd, check):
        seen["cmd"] = cmd
        seen["check"] = check
        return None

    monkeypatch.setattr(graph_create.subprocess, "run", _stub_run)
    graph_create.run_graph_sync_create(
        graph_create.GraphSyncCreateOptions(
            destination=tmp_path / "new-project",
        )
    )

    assert seen["check"] is True
    assert seen["cmd"][:3] == ["copier", "copy", "--UNSAFE"]
    assert seen["cmd"][-2] == "gh:wordlift/graph-sync-template"
    assert seen["cmd"][-1] == str(tmp_path / "new-project")


def test_graph_sync_create_missing_copier_binary(tmp_path: Path, monkeypatch) -> None:
    def _stub_run(_cmd, check):
        raise FileNotFoundError("copier")

    monkeypatch.setattr(graph_create.subprocess, "run", _stub_run)

    try:
        graph_create.run_graph_sync_create(
            graph_create.GraphSyncCreateOptions(destination=tmp_path / "new-project")
        )
        assert False, "Expected UsageError for missing copier executable"
    except UsageError as exc:
        assert "uv sync --extra dev" in str(exc)


def test_graph_sync_create_non_interactive_missing_answers_surfaces_clear_error(
    tmp_path: Path, monkeypatch
) -> None:
    seen = {}

    def _stub_run(_cmd, check):
        seen["cmd"] = _cmd
        raise subprocess.CalledProcessError(returncode=2, cmd=["copier"])

    monkeypatch.setattr(graph_create.subprocess, "run", _stub_run)

    try:
        graph_create.run_graph_sync_create(
            graph_create.GraphSyncCreateOptions(
                destination=tmp_path / "new-project",
                non_interactive=True,
            )
        )
        assert False, "Expected UsageError for non-interactive missing answers"
    except UsageError as exc:
        assert "--non-interactive mode" in str(exc)
        assert "Secret question requires a default value" in str(exc)
        assert "exit code: 2" in str(exc)

    assert "--non-interactive" not in seen["cmd"]
    assert "--defaults" in seen["cmd"]


def test_graph_sync_create_template_override_pass_through(tmp_path: Path, monkeypatch) -> None:
    seen = {}

    def _stub_run(cmd, check):
        seen["cmd"] = cmd
        return None

    monkeypatch.setattr(graph_create.subprocess, "run", _stub_run)
    graph_create.run_graph_sync_create(
        graph_create.GraphSyncCreateOptions(
            destination=tmp_path / "new-project",
            template="./local-template",
        )
    )

    assert seen["cmd"][-2] == "./local-template"


def test_graph_sync_create_vcs_ref_pass_through(tmp_path: Path, monkeypatch) -> None:
    seen = {}

    def _stub_run(cmd, check):
        seen["cmd"] = cmd
        return None

    monkeypatch.setattr(graph_create.subprocess, "run", _stub_run)
    graph_create.run_graph_sync_create(
        graph_create.GraphSyncCreateOptions(
            destination=tmp_path / "new-project",
            vcs_ref="v1.2.3",
        )
    )

    assert "--vcs-ref" in seen["cmd"]
    idx = seen["cmd"].index("--vcs-ref")
    assert seen["cmd"][idx + 1] == "v1.2.3"


def test_graph_sync_create_rejects_non_empty_destination_without_force(tmp_path: Path) -> None:
    destination = tmp_path / "existing"
    destination.mkdir(parents=True)
    (destination / "file.txt").write_text("x", encoding="utf-8")

    try:
        graph_create.run_graph_sync_create(
            graph_create.GraphSyncCreateOptions(destination=destination)
        )
        assert False, "Expected UsageError for non-empty destination"
    except UsageError as exc:
        assert "Destination is not empty" in str(exc)



"""Convert link group CSV to RDF and optionally apply changes via WordLift API."""

from __future__ import annotations

import csv
import json
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
from functools import partial
from urllib.parse import quote_plus

import requests
from rdflib import Graph, Literal, Namespace, URIRef
from rdflib.namespace import RDF, XSD
from requests.adapters import HTTPAdapter
from tqdm import tqdm
from urllib3.util.retry import Retry

SCHEMA = Namespace("http://schema.org/")
SEOVOC = Namespace("https://w3id.org/seovoc/")


@dataclass
class LinkGroupsOptions:
    input_file: str
    output_format: str = "turtle"
    apply: bool = False
    dry_run: bool = False
    api_key: str | None = None
    concurrency: int = 2
    retries: int = 3


def get_link_group_iri_and_id(target_link_iri: str) -> tuple[str, str]:
    link_group_iri = target_link_iri.rsplit("/", 1)[0]
    link_group_identifier = link_group_iri.rsplit("/", 1)[-1]
    return link_group_iri, link_group_identifier


def requests_retry_session(
    retries: int = 3,
    backoff_factor: float = 0.3,
    status_forcelist: tuple[int, ...] = (500, 502, 504),
    session: requests.Session | None = None,
) -> requests.Session:
    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


def delete_link_group(lg_iri: str, wordlift_key: str, session: requests.Session, dry_run: bool = False):
    delete_url = f"https://api.wordlift.io/entities?id={quote_plus(lg_iri)}&include_children=true"
    if dry_run:
        return f"curl -X DELETE -H 'Authorization: Key {wordlift_key}' '{delete_url}'"
    try:
        response = session.delete(delete_url, headers={"Authorization": f"Key {wordlift_key}"})
        response.raise_for_status()
        if response.status_code not in [200, 204]:
            return f"Unexpected status {response.status_code} for DELETE {lg_iri}"
    except requests.exceptions.RequestException as exc:
        return f"Error deleting {lg_iri}: {exc}"
    return None


def create_link_group(lg_item, wordlift_key: str, session: requests.Session, dry_run: bool = False):
    lg_iri, lg_graph = lg_item
    create_url = "https://api.wordlift.io/entities"
    ttl_data = lg_graph.serialize(format="turtle")
    if dry_run:
        escaped_ttl_data = ttl_data.replace("'", "'\\''")
        return (
            "curl -X PUT -H 'Authorization: Key {key}' -H 'Content-Type: text/turtle' "
            "--data-binary $'{data}' '{url}'"
        ).format(key=wordlift_key, data=escaped_ttl_data, url=create_url)
    try:
        headers = {"Authorization": f"Key {wordlift_key}", "Content-Type": "text/turtle"}
        response = session.put(create_url, data=ttl_data.encode("utf-8"), headers=headers)
        response.raise_for_status()
        if response.status_code not in [200, 201]:
            return f"Unexpected status {response.status_code} for CREATE {lg_iri}"
    except requests.exceptions.RequestException as exc:
        return f"Error creating {lg_iri}: {exc}"
    return None


def patch_webpage(page_item, wordlift_key: str, session: requests.Session, dry_run: bool = False):
    page_iri, lg_iri = page_item
    patch_url = f"https://api.wordlift.io/entities?id={quote_plus(page_iri)}"

    json_ld_graph = {"@id": str(page_iri), str(SEOVOC.hasLinkGroup): {"@id": str(lg_iri)}}
    value_string = json.dumps(json_ld_graph)

    patch_payload = [
        {
            "op": "add",
            "path": f"/{SEOVOC.hasLinkGroup}",
            "value": value_string,
        }
    ]

    if dry_run:
        data_raw = json.dumps(patch_payload)
        return (
            "curl -L -X PATCH '{url}' -H 'Content-Type: application/json-patch+json' "
            "-H 'Accept: application/ld+json' -H 'Authorization: Key {key}' "
            "--data-raw '{data}'"
        ).format(url=patch_url, key=wordlift_key, data=data_raw)

    try:
        headers = {
            "Content-Type": "application/json-patch+json",
            "Accept": "application/ld+json",
            "Authorization": f"Key {wordlift_key}",
        }
        response = session.patch(patch_url, json=patch_payload, headers=headers)
        response.raise_for_status()
    except requests.exceptions.RequestException as exc:
        return f"Error patching {page_iri} for LG {lg_iri}: {exc}"
    return None


def run_concurrently(worker_func, items, description: str, max_workers: int) -> None:
    items_list = list(items)
    if not items_list:
        return

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        future_to_item = {executor.submit(worker_func, item): item for item in items_list}
        with tqdm(total=len(items_list), desc=description) as pbar:
            for future in as_completed(future_to_item):
                result = future.result()
                if result:
                    if "curl" in result:
                        tqdm.write(f"\n{result}")
                    else:
                        tqdm.write(result)
                pbar.update(1)


def apply_changes(webpages, link_groups_data, wordlift_key, retries, concurrency, dry_run=False) -> None:
    print("\n--- Applying Changes ---")
    session = requests_retry_session(retries=retries)

    delete_worker = partial(delete_link_group, wordlift_key=wordlift_key, session=session, dry_run=dry_run)
    create_worker = partial(create_link_group, wordlift_key=wordlift_key, session=session, dry_run=dry_run)
    patch_worker = partial(patch_webpage, wordlift_key=wordlift_key, session=session, dry_run=dry_run)

    print("\n1. Deleting existing Link Groups...")
    run_concurrently(delete_worker, link_groups_data.keys(), "Deleting Link Groups", concurrency)

    print("\n2. Creating new Link Groups and Links...")
    run_concurrently(create_worker, link_groups_data.items(), "Creating Link Groups", concurrency)

    print("\n3. Patching WebPage entities...")
    patch_jobs = []
    for page_iri, page_data in webpages.items():
        for lg_iri in page_data["link_groups"]:
            patch_jobs.append((page_iri, lg_iri))

    run_concurrently(patch_worker, patch_jobs, "Patching WebPages  ", concurrency)


def convert_csv_to_rdf(input_file: str):
    g = Graph()
    g.bind("schema", SCHEMA)
    g.bind("seovoc", SEOVOC)
    g.bind("xsd", XSD)

    webpages = {}
    link_groups = {}
    link_groups_graphs = {}

    with open(input_file, "r", encoding="utf-8") as csvfile:
        reader = csv.DictReader(csvfile)
        rows = list(reader)

        for row in tqdm(rows, desc="Processing CSV rows "):
            source_iri = URIRef(row["source_iri"])
            target_link_iri = URIRef(row["target_link_iri"])
            target_web_page_iri = URIRef(row["target_web_page_iri"])
            target_headline = Literal(row["target_headline"])
            target_url = Literal(row["target_url"])
            target_position = Literal(row["target_position"], datatype=XSD.integer)
            target_score = Literal(row["target_score"], datatype=XSD.float)
            target_anchor_text = Literal(row["target_anchor_text"])

            if source_iri not in webpages:
                webpages[source_iri] = {"link_groups": set()}
                g.add((source_iri, RDF.type, SCHEMA.WebPage))

            link_group_iri_str, link_group_id = get_link_group_iri_and_id(row["target_link_iri"])
            link_group_iri = URIRef(link_group_iri_str)
            webpages[source_iri]["link_groups"].add(link_group_iri)

            if link_group_iri not in link_groups:
                link_groups[link_group_iri] = True
                g.add((link_group_iri, RDF.type, SEOVOC.LinkGroup))
                g.add((link_group_iri, SEOVOC.identifier, Literal(link_group_id)))
                g.add((link_group_iri, SEOVOC.isLinkGroupOf, source_iri))
                g.add((source_iri, SEOVOC.hasLinkGroup, link_group_iri))

                lg_graph = Graph()
                lg_graph.bind("schema", SCHEMA)
                lg_graph.bind("seovoc", SEOVOC)
                lg_graph.add((link_group_iri, RDF.type, SEOVOC.LinkGroup))
                lg_graph.add((link_group_iri, SEOVOC.identifier, Literal(link_group_id)))
                lg_graph.add((link_group_iri, SEOVOC.isLinkGroupOf, source_iri))
                link_groups_graphs[link_group_iri] = lg_graph

            lg_graph = link_groups_graphs[link_group_iri]

            g.add((target_link_iri, RDF.type, SEOVOC.Link))
            g.add((target_link_iri, SEOVOC.position, target_position))
            g.add((target_link_iri, SEOVOC.name, target_headline))
            g.add((target_link_iri, SEOVOC.weight, target_score))
            g.add((target_link_iri, SEOVOC.anchorText, target_anchor_text))
            g.add((target_link_iri, SEOVOC.anchorValue, target_url))
            g.add((target_link_iri, SEOVOC.anchorResource, target_web_page_iri))
            g.add((target_link_iri, SEOVOC.isLinkOf, link_group_iri))
            g.add((link_group_iri, SEOVOC.hasLink, target_link_iri))

            for p, o in g.predicate_objects(subject=target_link_iri):
                lg_graph.add((target_link_iri, p, o))
            lg_graph.add((link_group_iri, SEOVOC.hasLink, target_link_iri))

    return g, webpages, link_groups_graphs


def run(options: LinkGroupsOptions) -> str:
    full_graph, webpages_data, link_groups_api_data = convert_csv_to_rdf(options.input_file)

    if options.apply:
        if not options.api_key:
            raise RuntimeError("API key required when --apply is set.")
        apply_changes(
            webpages_data,
            link_groups_api_data,
            options.api_key,
            options.retries,
            options.concurrency,
            options.dry_run,
        )
        print("\n--- Done ---")
        return ""

    return full_graph.serialize(format=options.output_format)



"""CLI for structured data generation."""

from __future__ import annotations

import json
import os
from pathlib import Path

import typer

from worai.core.profile_loader import load_profile, resolve_profile_api_key
from wordlift_sdk.render import RenderOptions
try:
    from wordlift_sdk.validation import validate_jsonld_from_url
except ImportError:  # pragma: no cover - depends on installed wordlift-sdk version
    def validate_jsonld_from_url(*_args, **_kwargs):
        raise RuntimeError("validate_jsonld_from_url is unavailable in this wordlift-sdk version")
from worai.structured_data import (
    CreateRequest,
    CreateWorkflow,
    GenerateRequest,
    GenerateWorkflow,
)
from worai.commands.validate import _emit_validation_result
from worai.core.structured_data_inventory import (
    DEFAULT_BASE_URL as INVENTORY_DEFAULT_BASE_URL,
    InventoryOptions,
    run_inventory,
)
from worai.errors import UsageError

app = typer.Typer(add_completion=False, no_args_is_help=True)
validate_app = typer.Typer(add_completion=False, no_args_is_help=True)
DEFAULT_BASE_URL = "https://api.wordlift.io"


def _resolve_value(
    ctx: typer.Context,
    value: str | None,
    path: str,
    env_key: str | None,
) -> str | None:
    if value:
        return value
    if env_key and env_key in os.environ:
        return os.environ[env_key]
    try:
        profile, _, _ = load_profile(ctx)
    except ValueError:
        return None
    profile_settings = getattr(profile, "settings", {}) or {}
    underscore_key = path.replace(".", "_")
    if underscore_key in profile_settings:
        return profile_settings[underscore_key]
    if path in profile_settings:
        return profile_settings[path]
    return None


def _resolve_bool_value(
    ctx: typer.Context,
    value: bool | None,
    path: str,
    env_key: str | None = None,
) -> bool | None:
    if value is not None:
        return value
    raw = _resolve_value(ctx, None, path, env_key)
    if raw is None:
        return None
    if isinstance(raw, bool):
        return raw
    token = str(raw).strip().lower()
    if token in {"1", "true", "yes", "on"}:
        return True
    if token in {"0", "false", "no", "off"}:
        return False
    raise UsageError(f"Invalid boolean value for {path}: {raw}")


@validate_app.command(
    "page",
    no_args_is_help=True,
    help="Validate JSON-LD extracted from a webpage.",
)
def validate_page(
    url: str = typer.Argument(..., help="Webpage URL to extract JSON-LD from."),
    shape: list[str] | None = typer.Option(
        None,
        "--shape",
        "-s",
        help="Shape path or packaged shape name (default: all packaged shapes).",
        show_default=False,
    ),
    report_file: str | None = typer.Option(
        None,
        "--report-file",
        help="Write the SHACL report text to this file.",
    ),
    format: str = typer.Option(
        "pretty",
        "--format",
        help="Output format: pretty or raw.",
        show_default=True,
    ),
    color: bool = typer.Option(
        True,
        "--color/--no-color",
        help="Colorize pretty output.",
        show_default=True,
    ),
    list_shapes: bool = typer.Option(
        False,
        "--list-shapes",
        help="List available packaged shapes and exit.",
    ),
    headed: bool = typer.Option(False, "--headed", help="Run the browser with a visible UI."),
    timeout_ms: int = typer.Option(30000, "--timeout-ms", help="Timeout (ms) for page loads."),
    wait_until: str = typer.Option(
        "networkidle",
        "--wait-until",
        help="Playwright wait strategy.",
    ),
    ignore_https_errors: bool = typer.Option(
        False,
        "--ignore-https-errors",
        help="Ignore HTTPS errors when rendering the page.",
    ),
) -> None:
    if list_shapes:
        from wordlift_sdk.validation import list_shape_names

        for name in list_shape_names():
            typer.echo(name)
        return

    options = RenderOptions(
        url=url,
        headless=not headed,
        timeout_ms=timeout_ms,
        wait_until=wait_until,
        ignore_https_errors=ignore_https_errors,
    )
    try:
        result = validate_jsonld_from_url(url, shape_specs=shape, render_options=options)
    except Exception as exc:
        raise UsageError(
            "Failed to validate JSON-LD from webpage. "
            "Ensure the page renders and contains JSON-LD."
        ) from exc

    _emit_validation_result(result, format=format, color=color, report_file=report_file)


@app.command("create", help="Create YARRRML and JSON-LD for a webpage.")
def create(
    ctx: typer.Context,
    url: str = typer.Argument(..., help="Target page URL."),
    target_type_arg: str | None = typer.Argument(
        None, help="Schema.org type to generate (e.g., Review)."
    ),
    target_type: str | None = typer.Option(
        None, "--type", help="Schema.org type to generate (e.g., Review)."
    ),
    output_dir: Path = typer.Option(Path("."), "--output-dir", help="Output directory."),
    base_name: str = typer.Option("structured-data", "--base-name", help="Base output filename."),
    jsonld_path: Path | None = typer.Option(
        None, "--jsonld", help="Write JSON-LD to this file path."
    ),
    yarrml_path: Path | None = typer.Option(
        None, "--yarrml", help="Write YARRRML to this file path."
    ),
    debug: bool = typer.Option(False, "--debug", help="Write agent prompt/response to disk."),
    headed: bool = typer.Option(False, "--headed", help="Run the browser with a visible UI."),
    timeout_ms: int = typer.Option(30000, "--timeout-ms", help="Timeout (ms) for page loads."),
    max_retries: int = typer.Option(
        2, "--max-retries", help="Max retries for agent refinement when required props are missing."
    ),
    quality_check: bool = typer.Option(
        True,
        "--quality-check/--no-quality-check",
        help="Use agent-based quality scoring to decide retries.",
    ),
    max_xhtml_chars: int = typer.Option(
        40000, "--max-xhtml-chars", help="Max characters to keep in cleaned XHTML."
    ),
    max_text_node_chars: int = typer.Option(
        400, "--max-text-node-chars", help="Max characters per text node in cleaned XHTML."
    ),
    max_nesting_depth: int = typer.Option(
        2, "--max-nesting-depth", help="Max depth for related types in the schema guide."
    ),
    verbose: bool = typer.Option(
        True, "--verbose/--no-verbose", help="Emit progress logs to stderr."
    ),
    validate: bool = typer.Option(
        False, "--validate", help="Validate JSON-LD output with SHACL shapes."
    ),
    wait_until: str = typer.Option(
        "networkidle",
        "--wait-until",
        help="Playwright wait strategy.",
    ),
) -> None:
    if target_type is None:
        target_type = target_type_arg
    if not target_type:
        raise UsageError("Schema.org type is required. Pass it as an argument or via --type.")
    try:
        api_key, _, _ = resolve_profile_api_key(ctx)
    except ValueError as exc:
        raise UsageError(str(exc)) from exc

    def log(message: str) -> None:
        if verbose:
            typer.echo(message, err=True)

    request = CreateRequest(
        url=url,
        target_type=target_type,
        output_dir=output_dir,
        base_name=base_name,
        jsonld_path=jsonld_path,
        yarrml_path=yarrml_path,
        api_key=api_key,
        base_url=DEFAULT_BASE_URL,
        debug=debug,
        headed=headed,
        timeout_ms=timeout_ms,
        max_retries=max_retries,
        quality_check=quality_check,
        max_xhtml_chars=max_xhtml_chars,
        max_text_node_chars=max_text_node_chars,
        max_nesting_depth=max_nesting_depth,
        verbose=verbose,
        validate=validate,
        wait_until=wait_until,
    )
    result = CreateWorkflow().run(request, log)
    typer.echo(json.dumps(result.__dict__, indent=2))


@app.command("generate", help="Generate RDF from YARRRML over URL inputs.")
def generate(
    ctx: typer.Context,
    input_value: str = typer.Argument(..., metavar="INPUT", help="Sitemap URL/path or page URL."),
    yarrrml_path: Path = typer.Option(..., "--yarrrml", help="Path to YARRRML mapping file."),
    regex: str = typer.Option(".*", "--regex", help="Regex to filter URLs (matches full URL)."),
    output_dir: Path = typer.Option(Path("."), "--output-dir", help="Output directory."),
    output_format: str = typer.Option(
        "ttl", "--format", help="Output format: ttl, jsonld, rdf, nt, nq."
    ),
    concurrency: str = typer.Option(
        "auto", "--concurrency", help="Worker count or 'auto' to adapt to responses."
    ),
    headed: bool = typer.Option(False, "--headed", help="Run the browser with a visible UI."),
    timeout_ms: int = typer.Option(30000, "--timeout-ms", help="Timeout (ms) for page loads."),
    wait_until: str = typer.Option("networkidle", "--wait-until", help="Playwright wait strategy."),
    max_xhtml_chars: int = typer.Option(
        40000, "--max-xhtml-chars", help="Max characters to keep in cleaned XHTML."
    ),
    max_text_node_chars: int = typer.Option(
        400, "--max-text-node-chars", help="Max characters per text node in cleaned XHTML."
    ),
    max_pages: int | None = typer.Option(None, "--max-pages", help="Max pages to process."),
    verbose: bool = typer.Option(True, "--verbose/--no-verbose", help="Emit progress logs to stderr."),
) -> None:
    try:
        api_key, _, _ = resolve_profile_api_key(ctx)
    except ValueError as exc:
        raise UsageError(str(exc)) from exc

    def log(message: str) -> None:
        if verbose:
            typer.echo(message, err=True)

    request = GenerateRequest(
        input_value=input_value,
        yarrrml_path=yarrrml_path,
        regex=regex,
        output_dir=output_dir,
        output_format=output_format,
        concurrency=concurrency,
        api_key=api_key,
        base_url=DEFAULT_BASE_URL,
        headed=headed,
        timeout_ms=timeout_ms,
        wait_until=wait_until,
        max_xhtml_chars=max_xhtml_chars,
        max_text_node_chars=max_text_node_chars,
        max_pages=max_pages,
        verbose=verbose,
    )
    summary = GenerateWorkflow().run(request, log)
    typer.echo(json.dumps(summary, indent=2))


@app.command("inventory", help="Create a structured-data inventory from discovered URLs.")
def inventory(
    ctx: typer.Context,
    source: str = typer.Argument(
        ...,
        help="Sitemap URL/path, local URL list file, or Google Spreadsheet URL/ID.",
    ),
    sheet_name: str | None = typer.Option(
        None,
        "--sheet-name",
        help="Source sheet tab name when SOURCE is a Google Spreadsheet.",
    ),
    source_type: str | None = typer.Option(
        None,
        "--source-type",
        help="Optional source parser override (e.g., debug-cloud).",
    ),
    ingest_source: str | None = typer.Option(
        None,
        "--ingest-source",
        help="SDK ingestion source axis: auto|urls|sitemap|sheets|local.",
    ),
    ingest_loader: str | None = typer.Option(
        None,
        "--ingest-loader",
        help=(
            "SDK ingestion loader axis: "
            "auto|simple|proxy|playwright|premium_scraper|web_scrape_api|passthrough."
        ),
    ),
    ingest_passthrough_when_html: bool | None = typer.Option(
        None,
        "--ingest-passthrough-when-html/--no-ingest-passthrough-when-html",
        help="Prefer passthrough when source records include embedded HTML.",
    ),
    output: str | None = typer.Option(None, "--output", help="Write inventory to CSV."),
    destination_sheet_id: str | None = typer.Option(
        None,
        "--destination-sheet-id",
        help="Google spreadsheet ID where inventory should be written.",
    ),
    destination_sheet_name: str | None = typer.Option(
        None,
        "--destination-sheet-name",
        help="Destination sheet tab name to write inventory into.",
    ),
    client_secrets: str | None = typer.Option(
        None,
        "--client-secrets",
        help="OAuth2 client secrets JSON for Google Sheets (if re-auth is needed).",
    ),
    token: str | None = typer.Option(
        None,
        "--token",
        help="Path to OAuth token JSON used for Google APIs.",
    ),
    port: int = typer.Option(8080, "--port", help="Local redirect port for OAuth flow."),
    timeout: float = typer.Option(30.0, "--timeout", help="HTTP timeout for sitemap/page requests."),
    concurrency: str = typer.Option(
        "auto",
        "--concurrency",
        help="Worker count or 'auto' to adapt to responses.",
    ),
) -> None:
    try:
        api_key, _, _ = resolve_profile_api_key(ctx)
    except ValueError as exc:
        raise UsageError(str(exc)) from exc
    if not api_key:
        raise UsageError("WORDLIFT_API_KEY is required (or set profiles.<name>.api_key in config).")

    resolved_client_secrets = _resolve_value(
        ctx, client_secrets, "gsc.client_secrets", "GSC_CLIENT_SECRETS"
    )
    resolved_token = (
        _resolve_value(ctx, token, "oauth.token", "OAUTH_TOKEN")
        or _resolve_value(ctx, token, "gsc.token", "GSC_TOKEN")
        or _resolve_value(ctx, token, "ga.token", "GA_TOKEN")
        or "oauth_token.json"
    )
    resolved_ingest_source = _resolve_value(ctx, ingest_source, "ingest.source", None)
    resolved_ingest_loader = _resolve_value(ctx, ingest_loader, "ingest.loader", None)
    resolved_passthrough_when_html = _resolve_bool_value(
        ctx,
        ingest_passthrough_when_html,
        "ingest.passthrough_when_html",
        None,
    )
    options = InventoryOptions(
        source=source,
        api_key=api_key,
        base_url=INVENTORY_DEFAULT_BASE_URL,
        timeout=timeout,
        source_sheet_name=sheet_name,
        output=output,
        destination_spreadsheet_id=destination_sheet_id,
        destination_sheet_name=destination_sheet_name,
        client_secrets=resolved_client_secrets,
        token=resolved_token,
        port=port,
        concurrency=concurrency,
        ingest_source=resolved_ingest_source,
        ingest_loader=resolved_ingest_loader,
        ingest_passthrough_when_html=resolved_passthrough_when_html,
        source_type=source_type,
    )
    rows = run_inventory(options)
    typer.echo(
        json.dumps(
            {
                "total": len(rows),
                "output": output,
                "destination_sheet_id": destination_sheet_id,
                "destination_sheet_name": destination_sheet_name,
                "ingest_source": resolved_ingest_source,
                "ingest_loader": resolved_ingest_loader,
                "ingest_passthrough_when_html": resolved_passthrough_when_html,
                "source_type": source_type,
            },
            indent=2,
        )
    )


app.add_typer(validate_app, name="validate", help="Validate JSON-LD extracted from webpages.")



"""SEO Report CLI entry point."""

import os
from pathlib import Path
from typing import Optional

import typer
from jinja2 import Environment, FileSystemLoader

from worai.core.profile_loader import load_profile
from worai.seoreport.core import ReportOptions, generate_report_data

app = typer.Typer(add_completion=False, no_args_is_help=True)

def _resolve_value(
    ctx: typer.Context,
    value: str | None,
    path: str,
    env_key: str | None,
) -> str | None:
    if value:
        return value
    if env_key and env_key in os.environ:
        return os.environ[env_key]
    try:
        profile, _, _ = load_profile(ctx)
    except ValueError:
        return None
    profile_settings = getattr(profile, "settings", {}) or {}
    underscore_key = path.replace(".", "_")
    if underscore_key in profile_settings:
        return profile_settings[underscore_key]
    if path in profile_settings:
        return profile_settings[path]
    return None

@app.callback(invoke_without_command=True)
def run(
    ctx: typer.Context,
    site: Optional[str] = typer.Option(None, "--site", help="GSC property URL (or gsc.id in config)."),
    url_regex: Optional[str] = typer.Option(None, "--url-regex", help="Regex to filter URLs."),
    output: Optional[str] = typer.Option(None, "--output", help="Output file path."),
    format: str = typer.Option("markdown", "--format", help="Output format: markdown or html."),
    client_secrets: Optional[str] = typer.Option(None, "--client-secrets", help="OAuth2 client secrets JSON."),
    token: Optional[str] = typer.Option(None, "--token", help="Path to store OAuth2 token."),
    ga_id: Optional[str] = typer.Option(None, "--ga-id", help="GA4 property ID for Analytics section."),
    ga_client_secrets: Optional[str] = typer.Option(None, "--ga-client-secrets", help="OAuth2 client secrets JSON for GA4."),
    ga_token: Optional[str] = typer.Option(None, "--ga-token", help="Deprecated. Use --token for shared OAuth token."),
    ga_direct_share: float = typer.Option(0.3, "--ga-direct-share", help="Share of Direct traffic reallocated to AI (0-1)."),
    ga_source_regex: str = typer.Option("(chatgpt|openai|gemini|copilot|perplexity)", "--ga-source-regex", help="Regex to detect AI traffic from GA4 source."),
    ga_channel_group_name: str = typer.Option("AI Channel Group", "--ga-channel-group-name", help="GA4 custom channel group name for AI traffic."),
    ga_channel_group_id: Optional[str] = typer.Option(None, "--ga-channel-group-id", help="GA4 custom channel group ID for AI traffic."),
    gsc_max_workers: int = typer.Option(3, "--gsc-max-workers", help="Max parallel workers for GSC requests."),
    ga_max_workers: int = typer.Option(3, "--ga-max-workers", help="Max parallel workers for GA4 requests."),
    port: int = typer.Option(8080, "--port", help="Local redirect port for OAuth flow."),
    inspect_limit: int = typer.Option(0, "--inspect-limit", help="Number of top pages to inspect (0 skips, GSC quota)."),
) -> None:
    """Generate SEO performance report."""
    
    # Resolve credentials/config
    resolved_site = _resolve_value(ctx, site, "gsc.id", "GSC_ID")
    if not resolved_site:
        raise typer.BadParameter("--site is required (or set profiles.<name>.gsc.id / GSC_ID).")

    resolved_client_secrets = _resolve_value(
        ctx,
        client_secrets,
        "gsc.client_secrets",
        "GSC_CLIENT_SECRETS",
    )
    resolved_token = (
        _resolve_value(ctx, token, "oauth.token", "OAUTH_TOKEN")
        or _resolve_value(ctx, token, "gsc.token", "GSC_TOKEN")
        or _resolve_value(ctx, token, "ga.token", "GA_TOKEN")
        or "oauth_token.json"
    )
    resolved_ga_id = _resolve_value(ctx, ga_id, "ga.id", "GA_ID")
    resolved_ga_client_secrets = _resolve_value(
        ctx,
        ga_client_secrets,
        "ga.client_secrets",
        "GA_CLIENT_SECRETS",
    ) or resolved_client_secrets
    resolved_ga_token = _resolve_value(ctx, ga_token, "ga.token", "GA_TOKEN")
    if token and ga_token and token != ga_token:
        raise typer.BadParameter("--token and --ga-token must point to the same file (single-token mode).")
    if resolved_ga_token and resolved_ga_token != resolved_token:
        raise typer.BadParameter("--ga-token must match --token (single-token mode).")
    resolved_ga_token = resolved_token
    
    # Default output filename
    if not output:
        ext = "html" if format == "html" else "md"
        timestamp = generate_timestamp_str()
        output = f"seo_report_{timestamp}.{ext}"

    options = ReportOptions(
        site=resolved_site,
        url_regex=url_regex,
        client_secrets=resolved_client_secrets,
        token=resolved_token,
        ga_id=resolved_ga_id,
        ga_client_secrets=resolved_ga_client_secrets,
        ga_token=resolved_ga_token,
        ga_direct_share=ga_direct_share,
        ga_source_regex=ga_source_regex,
        ga_channel_group_name=ga_channel_group_name,
        ga_channel_group_id=ga_channel_group_id,
        gsc_max_workers=gsc_max_workers,
        ga_max_workers=ga_max_workers,
        port=port,
        output=output,
        format=format,
        inspect_limit=inspect_limit
    )

    typer.echo(f"Generating {format} report for {resolved_site}...")
    
    try:
        if options.ga_direct_share < 0 or options.ga_direct_share > 1:
            raise typer.BadParameter("--ga-direct-share must be between 0 and 1.")
        data = generate_report_data(options)
        
        # Render
        template_dir = Path(__file__).parent / "templates"
        env = Environment(loader=FileSystemLoader(template_dir))
        
        template_name = "report.html" if format == "html" else "report.md"
        template = env.get_template(template_name)
        
        rendered = template.render(data=data)
        
        with open(output, "w", encoding="utf-8") as f:
            f.write(rendered)
            
        typer.echo(f"Report saved to: {output}")
        
    except Exception as e:
        typer.echo(f"Error generating report: {e}", err=True)
        raise typer.Exit(code=1)

def generate_timestamp_str():
    import datetime
    return datetime.datetime.now().strftime("%Y%m%d_%H%M%S")



from __future__ import annotations

import requests
from typer.testing import CliRunner

from worai.commands import graph as graph_cmd
from worai.core.graph import property_delete as property_delete_core
from worai.errors import UsageError


class _FakeResponse:
    def __init__(self, status_code: int = 200, text: str = "ok") -> None:
        self.status_code = status_code
        self.text = text


def test_normalize_predicate_accepts_curie() -> None:
    assert (
        property_delete_core.normalize_predicate("seovoc:html")
        == "https://w3id.org/seovoc/html"
    )


def test_normalize_predicate_rejects_unknown_prefix() -> None:
    try:
        property_delete_core.normalize_predicate("foo:bar")
        assert False, "Expected ValueError for unknown prefix."
    except ValueError as exc:
        assert "Unknown CURIE prefix" in str(exc)


def test_normalize_predicate_validation_errors() -> None:
    for raw in ("", "   ", "seovoc", "seovoc:"):
        try:
            property_delete_core.normalize_predicate(raw)
            assert False, f"Expected ValueError for {raw!r}"
        except ValueError:
            pass


def test_normalize_predicate_accepts_full_iri() -> None:
    iri = "https://example.com/predicate"
    assert property_delete_core.normalize_predicate(iri) == iri


def test_fetch_candidate_iris_errors_on_bad_shape(monkeypatch) -> None:
    monkeypatch.setattr(
        property_delete_core,
        "graphql_request",
        lambda *_args, **_kwargs: {"data": "bad"},
    )
    try:
        property_delete_core.fetch_candidate_iris(
            endpoint="https://api.wordlift.io/graphql",
            api_key="wl_123",
            predicate_iri="https://w3id.org/seovoc/html",
        )
        assert False, "Expected RuntimeError for malformed GraphQL response."
    except RuntimeError as exc:
        assert "data.entities" in str(exc)


def test_fetch_candidate_iris_filters_invalid_entries(monkeypatch) -> None:
    monkeypatch.setattr(
        property_delete_core,
        "graphql_request",
        lambda *_args, **_kwargs: {
            "data": {
                "entities": [
                    "bad",
                    {"iri": "  ", "value": "x"},
                    {"iri": "https://example.com/e1", "value": ""},
                    {"iri": "https://example.com/e2", "value": "v"},
                ]
            }
        },
    )
    out = property_delete_core.fetch_candidate_iris(
        endpoint="https://api.wordlift.io/graphql",
        api_key="wl_123",
        predicate_iri="https://w3id.org/seovoc/html",
    )
    assert out == ["https://example.com/e2"]


def test_remove_property_once_handles_request_exception(monkeypatch) -> None:
    def _raise(*_args, **_kwargs):
        raise requests.RequestException("network boom")

    monkeypatch.setattr(property_delete_core.requests, "patch", _raise)
    ok, message = property_delete_core._remove_property_once(
        entities_endpoint="https://api.wordlift.io/entities",
        api_key="wl_123",
        iri="https://example.com/e1",
        predicate_iri="https://w3id.org/seovoc/html",
        timeout=5,
    )
    assert ok is False
    assert "network boom" in message


def test_remove_property_once_handles_http_failure(monkeypatch) -> None:
    monkeypatch.setattr(
        property_delete_core.requests,
        "patch",
        lambda *_args, **_kwargs: _FakeResponse(status_code=500, text="server error"),
    )
    ok, message = property_delete_core._remove_property_once(
        entities_endpoint="https://api.wordlift.io/entities",
        api_key="wl_123",
        iri="https://example.com/e1",
        predicate_iri="https://w3id.org/seovoc/html",
        timeout=5,
    )
    assert ok is False
    assert "HTTP 500" in message


def test_remove_property_with_retries_sleeps_on_success_rate_delay(monkeypatch) -> None:
    sleeps: list[float] = []
    monkeypatch.setattr(property_delete_core, "time", type("T", (), {"sleep": sleeps.append})())
    monkeypatch.setattr(
        property_delete_core,
        "_remove_property_once",
        lambda **_kwargs: (True, ""),
    )
    iri, ok, message = property_delete_core._remove_property_with_retries(
        entities_endpoint="https://api.wordlift.io/entities",
        api_key="wl_123",
        iri="https://example.com/e1",
        predicate_iri="https://w3id.org/seovoc/html",
        timeout=5,
        retries=2,
        rate_delay=0.25,
    )
    assert iri == "https://example.com/e1"
    assert ok is True
    assert message == ""
    assert sleeps == [0.25]


def test_remove_property_with_retries_returns_failure_after_retries(monkeypatch) -> None:
    sleeps: list[float] = []
    monkeypatch.setattr(property_delete_core, "time", type("T", (), {"sleep": sleeps.append})())
    monkeypatch.setattr(
        property_delete_core,
        "_remove_property_once",
        lambda **_kwargs: (False, "HTTP 429"),
    )
    iri, ok, message = property_delete_core._remove_property_with_retries(
        entities_endpoint="https://api.wordlift.io/entities",
        api_key="wl_123",
        iri="https://example.com/e1",
        predicate_iri="https://w3id.org/seovoc/html",
        timeout=5,
        retries=3,
        rate_delay=0.0,
    )
    assert iri == "https://example.com/e1"
    assert ok is False
    assert message == "HTTP 429"
    assert sleeps == [1, 1]


def test_run_property_delete_dry_run(monkeypatch) -> None:
    called: dict[str, object] = {}

    def _stub_graphql_request(_endpoint, _api_key, _query, timeout=60, include_private=False):
        called["include_private"] = include_private
        return {
            "data": {
                "entities": [
                    {"iri": "https://example.com/e1", "value": "<html>...</html>"},
                    {"iri": "https://example.com/e2", "value": "x"},
                    {"iri": "https://example.com/e3", "value": None},
                ]
            }
        }

    monkeypatch.setattr(property_delete_core, "graphql_request", _stub_graphql_request)
    options = property_delete_core.PropertyDeleteOptions(
        api_key="wl_123",
        predicate="seovoc:html",
        dry_run=True,
    )
    summary = property_delete_core.run_property_delete(options)

    assert summary.predicate == "https://w3id.org/seovoc/html"
    assert summary.candidates == 2
    assert summary.attempted == 0
    assert summary.removed == 0
    assert summary.failed == 0
    assert called["include_private"] is True


def test_run_property_delete_applies_patch(monkeypatch) -> None:
    calls: list[tuple[str, str]] = []
    headers_seen: list[dict[str, str] | None] = []

    def _stub_graphql_request(_endpoint, _api_key, _query, timeout=60, include_private=False):
        return {
            "data": {
                "entities": [
                    {"iri": "https://example.com/e1", "value": "x"},
                    {"iri": "https://example.com/e2", "value": "y"},
                ]
            }
        }

    def _stub_patch(url, headers=None, data=None, timeout=60):
        calls.append((url, data or ""))
        headers_seen.append(headers)
        return _FakeResponse(status_code=200, text="ok")

    monkeypatch.setattr(property_delete_core, "graphql_request", _stub_graphql_request)
    monkeypatch.setattr(property_delete_core.requests, "patch", _stub_patch)

    options = property_delete_core.PropertyDeleteOptions(
        api_key="wl_123",
        predicate="seovoc:html",
        dry_run=False,
        workers=1,
    )
    summary = property_delete_core.run_property_delete(options)

    assert summary.candidates == 2
    assert summary.attempted == 2
    assert summary.removed == 2
    assert summary.failed == 0
    assert len(calls) == 2
    assert '"path": "/https://w3id.org/seovoc/html"' in calls[0][1]
    assert headers_seen[0] is not None
    assert headers_seen[0]["X-include-Private"] == "true"


def test_run_property_delete_handles_limit_and_empty_candidates(monkeypatch) -> None:
    monkeypatch.setattr(
        property_delete_core,
        "graphql_request",
        lambda *_args, **_kwargs: {"data": {"entities": []}},
    )
    summary = property_delete_core.run_property_delete(
        property_delete_core.PropertyDeleteOptions(
            api_key="wl_123",
            predicate="seovoc:html",
            dry_run=False,
            limit=1,
        )
    )
    assert summary.candidates == 0
    assert summary.attempted == 0
    assert summary.removed == 0
    assert summary.failed == 0


def test_run_property_delete_counts_failures(monkeypatch) -> None:
    monkeypatch.setattr(
        property_delete_core,
        "graphql_request",
        lambda *_args, **_kwargs: {
            "data": {
                "entities": [
                    {"iri": "https://example.com/e1", "value": "x"},
                    {"iri": "https://example.com/e2", "value": "y"},
                ]
            }
        },
    )
    monkeypatch.setattr(
        property_delete_core,
        "_remove_property_with_retries",
        lambda **kwargs: (
            kwargs["iri"],
            kwargs["iri"].endswith("/e1"),
            "" if kwargs["iri"].endswith("/e1") else "fail",
        ),
    )
    summary = property_delete_core.run_property_delete(
        property_delete_core.PropertyDeleteOptions(
            api_key="wl_123",
            predicate="seovoc:html",
            dry_run=False,
            workers=1,
        )
    )
    assert summary.attempted == 2
    assert summary.removed == 1
    assert summary.failed == 1


def test_graph_property_delete_cli_calls_core(monkeypatch) -> None:
    seen: list[property_delete_core.PropertyDeleteOptions] = []

    def _stub_run(options):
        seen.append(options)
        if options.dry_run:
            return property_delete_core.PropertyDeleteSummary(
                predicate="https://w3id.org/seovoc/html",
                candidates=2,
                attempted=0,
                removed=0,
                failed=0,
            )
        return property_delete_core.PropertyDeleteSummary(
            predicate="https://w3id.org/seovoc/html",
            candidates=2,
            attempted=2,
            removed=2,
            failed=0,
        )

    monkeypatch.setattr(graph_cmd, "run_property_delete", _stub_run)
    monkeypatch.setattr(graph_cmd, "resolve_profile_api_key", lambda _ctx, explicit_profile=None: ("wl_123", explicit_profile or "default", "worai.toml"))

    runner = CliRunner()
    result = runner.invoke(
        graph_cmd.app,
        ["property", "delete", "seovoc:html", "--yes"],
    )

    assert result.exit_code == 0
    assert len(seen) == 2
    assert seen[0].dry_run is True
    assert seen[1].dry_run is False


def test_graph_property_delete_cli_dry_run_stops_after_first_call(monkeypatch) -> None:
    seen: list[property_delete_core.PropertyDeleteOptions] = []

    def _stub_run(options):
        seen.append(options)
        return property_delete_core.PropertyDeleteSummary(
            predicate="https://w3id.org/seovoc/html",
            candidates=3,
            attempted=0,
            removed=0,
            failed=0,
        )

    monkeypatch.setattr(graph_cmd, "run_property_delete", _stub_run)
    monkeypatch.setattr(graph_cmd, "resolve_profile_api_key", lambda _ctx, explicit_profile=None: ("wl_123", explicit_profile or "default", "worai.toml"))
    result = CliRunner().invoke(graph_cmd.app, ["property", "delete", "seovoc:html", "--dry-run"])
    assert result.exit_code == 0
    assert len(seen) == 1
    assert seen[0].dry_run is True


def test_graph_property_delete_cli_returns_when_no_candidates(monkeypatch) -> None:
    seen: list[property_delete_core.PropertyDeleteOptions] = []

    def _stub_run(options):
        seen.append(options)
        return property_delete_core.PropertyDeleteSummary(
            predicate="https://w3id.org/seovoc/html",
            candidates=0,
            attempted=0,
            removed=0,
            failed=0,
        )

    monkeypatch.setattr(graph_cmd, "run_property_delete", _stub_run)
    monkeypatch.setattr(graph_cmd, "resolve_profile_api_key", lambda _ctx, explicit_profile=None: ("wl_123", explicit_profile or "default", "worai.toml"))
    result = CliRunner().invoke(graph_cmd.app, ["property", "delete", "seovoc:html", "--yes"])
    assert result.exit_code == 0
    assert len(seen) == 1


def test_graph_property_delete_cli_surfaces_core_value_errors(monkeypatch) -> None:
    monkeypatch.setattr(graph_cmd, "resolve_profile_api_key", lambda _ctx, explicit_profile=None: ("wl_123", explicit_profile or "default", "worai.toml"))
    monkeypatch.setattr(graph_cmd, "run_property_delete", lambda _opts: (_ for _ in ()).throw(ValueError("bad")))
    result = CliRunner().invoke(graph_cmd.app, ["property", "delete", "seovoc:html", "--yes"])
    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)
    assert "bad" in str(result.exception)


def test_graph_property_delete_cli_prompts_confirmation_without_yes(monkeypatch) -> None:
    seen_confirm = {"called": False}
    seen_runs: list[property_delete_core.PropertyDeleteOptions] = []

    def _stub_run(options):
        seen_runs.append(options)
        if options.dry_run:
            return property_delete_core.PropertyDeleteSummary(
                predicate="https://w3id.org/seovoc/html",
                candidates=1,
                attempted=0,
                removed=0,
                failed=0,
            )
        return property_delete_core.PropertyDeleteSummary(
            predicate="https://w3id.org/seovoc/html",
            candidates=1,
            attempted=1,
            removed=1,
            failed=0,
        )

    monkeypatch.setattr(graph_cmd, "resolve_profile_api_key", lambda _ctx, explicit_profile=None: ("wl_123", explicit_profile or "default", "worai.toml"))
    monkeypatch.setattr(graph_cmd, "run_property_delete", _stub_run)

    def _stub_confirm(*_args, **_kwargs):
        seen_confirm["called"] = True
        return True

    monkeypatch.setattr(graph_cmd.typer, "confirm", _stub_confirm)
    result = CliRunner().invoke(graph_cmd.app, ["property", "delete", "seovoc:html"])
    assert result.exit_code == 0
    assert seen_confirm["called"] is True
    assert len(seen_runs) == 2


def test_graph_property_delete_cli_raises_when_failures_present(monkeypatch) -> None:
    calls = {"count": 0}

    def _stub_run(options):
        calls["count"] += 1
        if options.dry_run:
            return property_delete_core.PropertyDeleteSummary(
                predicate="https://w3id.org/seovoc/html",
                candidates=1,
                attempted=0,
                removed=0,
                failed=0,
            )
        return property_delete_core.PropertyDeleteSummary(
            predicate="https://w3id.org/seovoc/html",
            candidates=1,
            attempted=1,
            removed=0,
            failed=1,
        )

    monkeypatch.setattr(graph_cmd, "resolve_profile_api_key", lambda _ctx, explicit_profile=None: ("wl_123", explicit_profile or "default", "worai.toml"))
    monkeypatch.setattr(graph_cmd, "run_property_delete", _stub_run)
    result = CliRunner().invoke(graph_cmd.app, ["property", "delete", "seovoc:html", "--yes"])
    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)
    assert "Some entities failed to patch" in str(result.exception)
    assert calls["count"] == 2


def test_graph_sync_create_prompts_for_destination(monkeypatch, tmp_path) -> None:
    captured: dict[str, object] = {}

    def _stub_prompt(_text: str) -> str:
        return str(tmp_path / "from-prompt")

    def _stub_create(options):
        captured["destination"] = options.destination

    monkeypatch.setattr(graph_cmd.typer, "prompt", _stub_prompt)
    monkeypatch.setattr(graph_cmd, "run_graph_sync_create", _stub_create)
    graph_cmd.create(
        destination=None,
        template="gh:wordlift/graph-sync-template",
        defaults=False,
        data_file=None,
        vcs_ref=None,
        non_interactive=False,
        force=False,
    )
    assert str(captured["destination"]).endswith("from-prompt")


def test_graph_property_delete_cli_requires_api_key(monkeypatch) -> None:
    monkeypatch.setattr(graph_cmd, "resolve_profile_api_key", lambda _ctx, explicit_profile=None: (None, explicit_profile or "default", "worai.toml"))
    runner = CliRunner()
    result = runner.invoke(graph_cmd.app, ["property", "delete", "seovoc:html", "--yes"])

    assert result.exit_code == 1
    assert isinstance(result.exception, UsageError)
    assert "WORDLIFT_API_KEY is required" in str(result.exception)



"""CLI wrapper for graph workflows."""

from __future__ import annotations

from datetime import datetime
from pathlib import Path

import typer

from worai.config import resolve_config_path
from worai.core.graph import (
    GraphExportOptions,
    GraphSyncCreateOptions,
    GraphSyncOptions,
    PropertyDeleteOptions,
    run_graph_export,
    run_graph_sync,
    run_graph_sync_create,
    run_property_delete,
    validate_exported_graph,
)
from worai.core.profile_loader import resolve_profile_api_key
from worai.errors import UsageError

app = typer.Typer(add_completion=False, no_args_is_help=True)
sync_app = typer.Typer(add_completion=False, no_args_is_help=True)
property_app = typer.Typer(add_completion=False, no_args_is_help=True)


@sync_app.command(
    "run",
    no_args_is_help=True,
    help="Run graph sync for a configured profile.",
)
def run(
    ctx: typer.Context,
    profile: str = typer.Option(..., "--profile", help="Profile name in worai.toml."),
    debug: bool = typer.Option(
        False,
        "--debug",
        help="Write generated callback graphs to output/debug_cloud/<profile>/.",
    ),
) -> None:
    try:
        config_path = (ctx.obj or {}).get("config_path") if ctx else None
        run_graph_sync(
            GraphSyncOptions(
                profile=profile,
                config_path=config_path or resolve_config_path(None),
                debug=debug,
            )
        )
    except ValueError as exc:
        raise UsageError(str(exc)) from exc


@sync_app.command(
    "create",
    no_args_is_help=True,
    help="Create a graph sync project from a Copier template.",
)
def create(
    destination: Path | None = typer.Argument(
        None,
        metavar="DESTINATION",
        help="Output directory where the graph sync project will be created.",
    ),
    template: str = typer.Option(
        "gh:wordlift/graph-sync-template",
        "--template",
        help="Copier template source (Git URL, gh: shortcut, or local path).",
    ),
    defaults: bool = typer.Option(
        False,
        "--defaults",
        help="Use Copier default answers.",
    ),
    data_file: Path | None = typer.Option(
        None,
        "--data-file",
        help="Path to a Copier data file with pre-filled answers.",
    ),
    vcs_ref: str | None = typer.Option(
        None,
        "--vcs-ref",
        help="Template VCS ref (tag, branch, or commit).",
    ),
    non_interactive: bool = typer.Option(
        False,
        "--non-interactive",
        help="Disable prompts and fail if required answers are missing.",
    ),
    force: bool = typer.Option(
        False,
        "--force",
        help="Allow writing into existing destinations.",
    ),
) -> None:
    if destination is None:
        if non_interactive:
            raise UsageError("Destination path is required when using --non-interactive.")
        destination = Path(typer.prompt("Destination directory"))

    run_graph_sync_create(
        GraphSyncCreateOptions(
            destination=destination,
            template=template,
            defaults=defaults,
            data_file=data_file,
            vcs_ref=vcs_ref,
            non_interactive=non_interactive,
            force=force,
        )
    )


app.add_typer(sync_app, name="sync", help="Run graph sync workflows.")


def _default_export_output_path(profile: str) -> Path:
    day = datetime.now().strftime("%Y%m%d")
    sequence = 1
    while True:
        candidate = Path(f"export_{profile}_{day}_{sequence}.ttl")
        if not candidate.exists():
            return candidate
        sequence += 1


@app.command(
    "export",
    no_args_is_help=False,
    help="Export graph data from the WordLift /dataset/export endpoint.",
)
def export(
    ctx: typer.Context,
    output_file_name: str | None = typer.Argument(
        None,
        metavar="OUTPUT_FILE_NAME",
        help="Output file path. Extension controls export format (.ttl, .nt, .nq, .rdf, .xml, .jsonld, .json).",
    ),
    profile: str = typer.Option(
        "default",
        "--profile",
        help="Profile name in worai.toml. Defaults to 'default'.",
    ),
    endpoint: str = typer.Option(
        "https://api.wordlift.io/dataset/export",
        "--endpoint",
        help="WordLift export endpoint.",
    ),
    timeout: int = typer.Option(120, "--timeout", help="HTTP timeout in seconds."),
    validate: bool = typer.Option(
        False,
        "--validate",
        help="Run SHACL validation on the exported graph file.",
    ),
) -> None:
    try:
        api_key, resolved_profile, _ = resolve_profile_api_key(ctx, explicit_profile=profile)
    except ValueError as exc:
        raise UsageError(str(exc)) from exc
    if not api_key:
        raise UsageError(
            f"WORDLIFT_API_KEY is required (or set profiles.{resolved_profile}.api_key in config)."
        )

    output_path = Path(output_file_name) if output_file_name else _default_export_output_path(profile)
    try:
        exported = run_graph_export(
            GraphExportOptions(
                api_key=api_key,
                output_path=output_path,
                endpoint=endpoint,
                timeout=timeout,
            )
        )
    except ValueError as exc:
        raise UsageError(str(exc)) from exc
    except RuntimeError as exc:
        raise UsageError(str(exc)) from exc

    if validate:
        try:
            validate_exported_graph(exported)
        except RuntimeError as exc:
            raise UsageError(str(exc)) from exc
        typer.echo(f"Validated graph export: {exported}")

    typer.echo(f"Exported graph to {exported}")


@property_app.command(
    "delete",
    no_args_is_help=True,
    help="Delete one predicate from all matching entities in the graph.",
)
def property_delete(
    ctx: typer.Context,
    predicate: str = typer.Argument(..., help="Predicate IRI or CURIE (for example, seovoc:html)."),
    endpoint: str = typer.Option(
        "https://api.wordlift.io/graphql", "--endpoint", help="WordLift GraphQL endpoint."
    ),
    entities_endpoint: str = typer.Option(
        "https://api.wordlift.io/entities", "--entities-endpoint", help="WordLift entities API endpoint."
    ),
    dry_run: bool = typer.Option(False, "--dry-run", help="List matches without applying changes."),
    yes: bool = typer.Option(False, "--yes", help="Skip confirmation prompt."),
    workers: int = typer.Option(4, "--workers", help="Number of concurrent PATCH requests."),
    retries: int = typer.Option(3, "--retries", help="Retries per failed PATCH request."),
    rate_delay: float = typer.Option(0.0, "--rate-delay", help="Delay in seconds after each successful PATCH."),
    limit: int = typer.Option(0, "--limit", help="Limit number of matching entities to process."),
    timeout: int = typer.Option(60, "--timeout", help="HTTP timeout in seconds."),
) -> None:
    try:
        api_key, _, _ = resolve_profile_api_key(ctx)
    except ValueError as exc:
        raise UsageError(str(exc)) from exc
    if not api_key:
        raise UsageError("WORDLIFT_API_KEY is required (or set profiles.<name>.api_key in config).")

    options = PropertyDeleteOptions(
        api_key=api_key,
        predicate=predicate,
        graphql_endpoint=endpoint,
        entities_endpoint=entities_endpoint,
        dry_run=True,
        workers=workers,
        retries=retries,
        rate_delay=rate_delay,
        limit=limit,
        timeout=timeout,
    )
    try:
        summary = run_property_delete(options)
    except ValueError as exc:
        raise UsageError(str(exc)) from exc

    typer.echo(f"Predicate: {summary.predicate}")
    typer.echo(f"Matching entities: {summary.candidates}")
    if dry_run:
        return

    if summary.candidates == 0:
        return

    if not yes:
        typer.confirm(
            f"Delete predicate {summary.predicate} from {summary.candidates} entities?",
            abort=True,
        )

    execute_options = PropertyDeleteOptions(
        api_key=options.api_key,
        predicate=options.predicate,
        graphql_endpoint=options.graphql_endpoint,
        entities_endpoint=options.entities_endpoint,
        dry_run=False,
        workers=options.workers,
        retries=options.retries,
        rate_delay=options.rate_delay,
        limit=options.limit,
        timeout=options.timeout,
    )
    result = run_property_delete(execute_options)
    typer.echo(f"Attempted: {result.attempted}")
    typer.echo(f"Removed: {result.removed}")
    typer.echo(f"Failed: {result.failed}")
    if result.failed:
        raise UsageError("Some entities failed to patch. Re-run with lower --workers or higher --retries.")


app.add_typer(property_app, name="property", help="Bulk predicate operations on the graph.")



from __future__ import annotations

import json
from pathlib import Path

import pytest

from worai.core import canonicalize_duplicate_pages as canon
from worai.core import dedupe
from worai.core import link_groups


def test_canonicalize_helpers_and_run(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    assert canon.normalize_name("  A   B ") == "a b"
    assert canon.parse_float("3.2") == 3.2
    assert canon.parse_float("x") == 0.0

    query = canon.build_entity_query(["https://example.com/a"])
    assert "entity(url" in query

    types = canon.extract_types([{"iri": "http://schema.org/Product"}, "Other"])
    assert "http://schema.org/Product" in types
    assert canon.choose_primary_type(["b", "a"]) == "a"

    label, matcher = canon.build_type_matcher("schema:Product")
    assert label == "Product"
    assert matcher is not None and matcher(["http://schema.org/Product"])

    selected = canon.select_entity_for_url(
        "u",
        [{"name": "n", "types": ["http://schema.org/Product"]}],
        "Product",
        matcher,
    )
    assert selected is not None

    with pytest.raises(Exception):
        canon.select_entity_for_url("u", [{"name": "a"}, {"name": "b"}], None, None)

    clusters = canon.build_clusters(
        {"u1": {"name": "Product A", "types": ["http://schema.org/Product"]}},
        "Product",
    )
    assert len(clusters) == 1

    out = tmp_path / "out.csv"
    canon.write_output_csv(str(out), [])
    assert out.read_text().startswith("type,name")

    in_csv = tmp_path / "gsc.csv"
    in_csv.write_text("page,clicks_28d\nhttps://example.com/a,10\nhttps://example.com/b,5\n", encoding="utf-8")
    monkeypatch.setattr(
        canon,
        "fetch_entities_by_url",
        lambda *_a, **_k: {
            "https://example.com/a": {"name": "X", "types": ["http://schema.org/Product"]},
            "https://example.com/b": {"name": "X", "types": ["http://schema.org/Product"]},
        },
    )
    opts = canon.CanonicalizeOptions(
        api_key="wl",
        input_csv=str(in_csv),
        output_csv=str(out),
        url_regex=None,
        entity_type="Product",
    )
    canon.run(opts)
    assert "duplicate_url" in out.read_text()


def test_fetch_entities_by_url_uses_graphql(monkeypatch: pytest.MonkeyPatch) -> None:
    def _stub(_endpoint, _key, _query):
        return {"data": {"e0": {"iri": "i1"}}}

    monkeypatch.setattr(canon, "graphql_request", _stub)
    out = canon.fetch_entities_by_url("e", "k", ["https://example.com/a"], batch_size=1)
    assert out["https://example.com/a"]["iri"] == "i1"


def test_dedupe_fetch_prompt_delete_and_run(monkeypatch: pytest.MonkeyPatch) -> None:
    class _Resp:
        status_code = 200
        text = "ok"

        @staticmethod
        def json():
            return {"data": {"entities": [{"iri": "i1", "url": "u"}, {"iri": "i2", "url": "u"}]}}

    monkeypatch.setattr(dedupe.requests, "post", lambda *_a, **_k: _Resp())
    entities = dedupe.fetch_entities("e", "k")
    assert len(entities) == 2

    inputs = iter(["s", "1"])
    monkeypatch.setattr("builtins.input", lambda _prompt: next(inputs))
    idx, skipped = dedupe.prompt_choice("u", ["i1", "i2"])
    assert skipped
    idx, skipped = dedupe.prompt_choice("u", ["i1", "i2"])
    assert idx == 0 and not skipped

    monkeypatch.setattr(dedupe.requests, "delete", lambda *_a, **_k: type("R", (), {"status_code": 204, "text": ""})())
    monkeypatch.setattr(dedupe.time, "sleep", lambda _x: None)
    assert dedupe.delete_iri("i1", "k", dry_run=False, rate_delay=0.1)

    monkeypatch.setattr(dedupe, "fetch_entities", lambda *_a, **_k: [{"iri": "i1", "url": "u"}, {"iri": "i2", "url": "u"}])
    monkeypatch.setattr(dedupe, "prompt_choice", lambda *_a, **_k: (0, False))
    monkeypatch.setattr(dedupe, "delete_iri", lambda *_a, **_k: True)
    dedupe.run(dedupe.DedupeOptions(api_key="k"))


def test_link_groups_helpers(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    iri, ident = link_groups.get_link_group_iri_and_id("https://example.com/lg/1/link/2")
    assert iri.endswith("/1/link")
    assert ident == "link"

    sess = link_groups.requests_retry_session()
    assert hasattr(sess, "mount")

    class _Session:
        def delete(self, *_a, **_k):
            return type("R", (), {"status_code": 204, "raise_for_status": lambda self: None})()

        def put(self, *_a, **_k):
            return type("R", (), {"status_code": 201, "raise_for_status": lambda self: None})()

        def patch(self, *_a, **_k):
            return type("R", (), {"status_code": 200, "raise_for_status": lambda self: None})()

    assert "curl -X DELETE" in link_groups.delete_link_group("https://example.com/lg/1", "k", _Session(), dry_run=True)
    assert link_groups.delete_link_group("https://example.com/lg/1", "k", _Session(), dry_run=False) is None

    g = link_groups.Graph()
    lg_iri = link_groups.URIRef("https://example.com/lg/1")
    assert "curl -X PUT" in link_groups.create_link_group((lg_iri, g), "k", _Session(), dry_run=True)
    assert link_groups.create_link_group((lg_iri, g), "k", _Session(), dry_run=False) is None

    assert "curl -L -X PATCH" in link_groups.patch_webpage(("https://example.com/p", "https://example.com/lg/1"), "k", _Session(), dry_run=True)
    assert link_groups.patch_webpage(("https://example.com/p", "https://example.com/lg/1"), "k", _Session(), dry_run=False) is None

    link_groups.run_concurrently(lambda _i: None, [1, 2], "x", 2)

    csv_path = tmp_path / "links.csv"
    csv_path.write_text(
        "source_iri,target_link_iri,target_web_page_iri,target_headline,target_url,target_position,target_score,target_anchor_text\n"
        "https://example.com/p,https://example.com/lg/1/l1,https://example.com/t,Headline,https://example.com/t,1,0.5,anchor\n",
        encoding="utf-8",
    )
    full_graph, webpages, api_graphs = link_groups.convert_csv_to_rdf(str(csv_path))
    assert len(full_graph) > 0
    assert len(webpages) == 1
    assert len(api_graphs) == 1

    monkeypatch.setattr(link_groups, "convert_csv_to_rdf", lambda _f: (full_graph, webpages, api_graphs))
    monkeypatch.setattr(link_groups, "apply_changes", lambda *_a, **_k: None)
    assert link_groups.run(link_groups.LinkGroupsOptions(input_file=str(csv_path), output_format="turtle"))
    assert link_groups.run(link_groups.LinkGroupsOptions(input_file=str(csv_path), apply=True, api_key="k")) == ""



from __future__ import annotations

from pathlib import Path

import pytest
from rdflib import BNode, Graph, Literal, RDF, URIRef
from rdflib.collection import Collection
from typer.testing import CliRunner

from worai.commands import validate as validate_cmd
from worai.errors import UsageError


class _Result:
    def __init__(self, conforms: bool = True, warnings: int = 0) -> None:
        self.conforms = conforms
        self.warning_count = warnings
        self.report_text = "REPORT"
        self.report_graph = Graph()
        self.data_graph = Graph()
        self.shape_source_map = {}


def test_term_and_path_helpers() -> None:
    g = Graph()
    p1 = URIRef("http://schema.org/name")
    p2 = URIRef("http://example.com/ns#value")
    chain = BNode()
    Collection(g, chain, [p1, p2])

    assert validate_cmd._shorten_term(p1) == "name"
    assert validate_cmd._path_to_str(g, p1) == "name"
    assert validate_cmd._path_to_str(g, chain) == "name.value"
    assert validate_cmd._describe_path(g, p2) == "value"

    inv = BNode()
    g.add((inv, validate_cmd._SH.inversePath, p1))
    assert "inverse name" == validate_cmd._describe_path(g, inv)


def test_or_shape_hint_and_pretty_lines() -> None:
    g = Graph()
    result = URIRef("urn:result")
    focus = URIRef("https://example.com/page")
    shape = URIRef("urn:shape/google_required")
    prop = BNode()
    alt = BNode()
    or_list = BNode()
    path = URIRef("http://schema.org/headline")
    g.add((result, validate_cmd._SH.resultSeverity, validate_cmd._SH.Violation))
    g.add((result, validate_cmd._SH.focusNode, focus))
    g.add((result, validate_cmd._SH.resultPath, path))
    g.add((result, validate_cmd._SH.resultMessage, Literal("Less than 1 values")))
    g.add((result, validate_cmd._SH.sourceShape, shape))
    g.add((shape, validate_cmd._SH["or"], or_list))
    Collection(g, or_list, [alt])
    g.add((alt, validate_cmd._SH.property, prop))
    g.add((prop, validate_cmd._SH.path, path))
    data = Graph()
    data.add((focus, RDF.type, URIRef("http://schema.org/Thing")))

    hint = validate_cmd._or_shape_hint(g, shape)
    assert hint is not None and "headline" in hint
    header, lines = validate_cmd._format_pretty_lines(g, data, {shape: "shape.ttl"})
    assert header.startswith("Status: ERROR")
    assert any("Missing required property" in line for _label, line in lines)


def test_emit_validation_result_modes(tmp_path: Path) -> None:
    ok = _Result(conforms=True, warnings=0)
    validate_cmd._emit_validation_result(ok, format="raw", color=False, report_file=str(tmp_path / "raw.txt"))
    assert (tmp_path / "raw.txt").read_text() == "REPORT"

    with pytest.raises(UsageError, match="Invalid format"):
        validate_cmd._emit_validation_result(ok, format="bad", color=False, report_file=None)

    bad = _Result(conforms=False, warnings=0)
    with pytest.raises(UsageError, match="validation failed"):
        validate_cmd._emit_validation_result(bad, format="pretty", color=False, report_file=None)

    warn = _Result(conforms=True, warnings=1)
    with pytest.raises(UsageError, match="warnings found"):
        validate_cmd._emit_validation_result(warn, format="pretty", color=False, report_file=None)


def test_validate_jsonld_input_url_error(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(validate_cmd, "validate_file", lambda *_a, **_k: (_ for _ in ()).throw(RuntimeError("x")))
    with pytest.raises(UsageError, match="structured-data validate page"):
        validate_cmd._validate_jsonld_input("https://example.com/page", None)


def test_main_list_shapes_and_missing_input(monkeypatch: pytest.MonkeyPatch) -> None:
    runner = CliRunner()
    monkeypatch.setattr(validate_cmd, "_list_shapes", lambda: None)
    result = runner.invoke(validate_cmd.app, ["--list-shapes"])
    assert result.exit_code == 0

    result = runner.invoke(validate_cmd.app, [])
    assert result.exit_code == 2



"""Graph sync workflow helpers."""

from worai.core.graph.create import GraphSyncCreateOptions, run_graph_sync_create
from worai.core.graph.export import GraphExportOptions, run_graph_export, validate_exported_graph
from worai.core.graph.property_delete import PropertyDeleteOptions, run_property_delete
from worai.core.graph.sync import GraphSyncOptions, run_graph_sync

__all__ = [
    "GraphExportOptions",
    "GraphSyncCreateOptions",
    "GraphSyncOptions",
    "PropertyDeleteOptions",
    "run_graph_export",
    "validate_exported_graph",
    "run_graph_sync",
    "run_graph_sync_create",
    "run_property_delete",
]



from __future__ import annotations

import json
from pathlib import Path
from types import SimpleNamespace

import pytest

from worai.seocheck import cli as mod
from worai.seocheck.checks.types import CheckResult


def _site_result(status: str = "ok"):
    return SimpleNamespace(url="u", status=status, details="d", status_code=200, disallow_all=False)


def _llms_result(status: str = "ok"):
    return SimpleNamespace(url="l", status=status, details="d", status_code=200)


def test_json_print_helpers_and_assets(tmp_path: Path, capsys: pytest.CaptureFixture[str]) -> None:
    mod._print_site_results([{"name": "robots", "status": "ok", "details": "d", "url": "u"}], as_json=True)
    j1 = json.loads(capsys.readouterr().out.strip())
    assert j1["name"] == "robots"

    mod._print_page_results("u", [CheckResult(name="c", status="ok")], as_json=True)
    j2 = json.loads(capsys.readouterr().out.strip())
    assert j2["type"] == "page"

    mod._copy_report_assets(tmp_path)
    assert (tmp_path / "index.html").exists()


def test_serve_report_handles_keyboard_interrupt(monkeypatch: pytest.MonkeyPatch, tmp_path: Path, capsys: pytest.CaptureFixture[str]) -> None:
    class _Server:
        server_address = ("127.0.0.1", 1234)

        def serve_forever(self):
            raise KeyboardInterrupt()

        def shutdown(self):
            print("shutdown")

    monkeypatch.setattr(mod.http.server, "ThreadingHTTPServer", lambda *_a, **_k: _Server())
    opened = {}
    monkeypatch.setattr(mod.webbrowser, "open", lambda url: opened.setdefault("url", url))
    mod._serve_report(tmp_path, open_browser=True)
    out = capsys.readouterr().out
    assert "Report server running" in out
    assert "shutdown" in out
    assert opened["url"].startswith("http://127.0.0.1:1234/")


def test_run_invalid_concurrency_and_unknown_checks(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(mod, "parse_sitemap_urls", lambda *_a, **_k: ["https://e.com/p"])
    monkeypatch.setattr(mod, "check_robots_txt", lambda *_a, **_k: _site_result())
    monkeypatch.setattr(mod, "check_llms_txt", lambda *_a, **_k: _llms_result())
    monkeypatch.setattr(mod, "get_page_checks", lambda **_k: [SimpleNamespace(name="status")])

    assert mod.run(["https://e.com/sitemap.xml", "--concurrency", "x", "--no-report-ui"]) == 1
    assert mod.run(["https://e.com/sitemap.xml", "--checks", "missing", "--no-report-ui"]) == 1
    assert mod.run(["https://e.com/sitemap.xml", "--disable-checks", "missing", "--no-report-ui"]) == 1


def test_run_recheck_and_merge_and_outputs(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    report = tmp_path / "report.json"
    report.write_text(json.dumps({"pages": [{"url": "https://e.com/a", "status": "fail", "checks": []}, {"url": "https://e.com/old", "status": "ok", "checks": []}]}), encoding="utf-8")

    monkeypatch.setattr(mod, "parse_sitemap_urls", lambda *_a, **_k: ["https://e.com/a", "https://e.com/b"])
    monkeypatch.setattr(mod, "check_robots_txt", lambda *_a, **_k: _site_result())
    monkeypatch.setattr(mod, "check_llms_txt", lambda *_a, **_k: _llms_result())
    monkeypatch.setattr(mod, "get_page_checks", lambda **_k: [SimpleNamespace(name="status")])

    def _proc(url: str, **_k):
        return (
            {"type": "page", "url": url, "checks": [{"name": "status", "status": "ok"}]},
            {"url": url, "status": "ok", "file": None, "html": None, "checks": [{"name": "status", "status": "ok"}]},
            False,
        )

    monkeypatch.setattr(mod, "_process_url", _proc)
    served = {}
    monkeypatch.setattr(mod, "_serve_report", lambda output_dir, open_browser: served.setdefault("called", (output_dir, open_browser)))

    rc = mod.run([
        "https://e.com/sitemap.xml",
        "--recheck-failed",
        "--recheck-from",
        str(report),
        "--concurrency",
        "2",
        "--output",
        str(tmp_path / "o.json"),
        "--output-summary",
        str(tmp_path / "s.txt"),
    ])
    assert rc == 0
    payload = json.loads((tmp_path / "o.json").read_text(encoding="utf-8"))
    urls = [p["url"] for p in payload["pages"]]
    assert "https://e.com/old" in urls
    assert served.get("called")


def test_run_auto_concurrency_and_recheck_shortcuts(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.setattr(mod, "check_robots_txt", lambda *_a, **_k: _site_result())
    monkeypatch.setattr(mod, "check_llms_txt", lambda *_a, **_k: _llms_result())
    monkeypatch.setattr(mod, "get_page_checks", lambda **_k: [SimpleNamespace(name="status")])

    assert mod.run(["https://e.com/sitemap.xml", "--recheck-failed", "--no-report-ui"]) == 1

    rp = tmp_path / "r.json"
    rp.write_text(json.dumps({"pages": [{"url": "https://e.com/x", "status": "ok", "checks": []}]}), encoding="utf-8")
    monkeypatch.setattr(mod, "parse_sitemap_urls", lambda *_a, **_k: ["https://e.com/x"])
    assert mod.run(["https://e.com/sitemap.xml", "--recheck-failed", "--recheck-from", str(rp), "--no-report-ui"]) == 0

    rp.write_text(json.dumps({"pages": [{"url": "https://e.com/x", "status": "fail", "checks": []}]}), encoding="utf-8")
    monkeypatch.setattr(mod, "parse_sitemap_urls", lambda *_a, **_k: ["https://e.com/a", "https://e.com/b", "https://e.com/c"])

    calls = {"n": 0}

    def _proc(url: str, **_k):
        calls["n"] += 1
        had_error = calls["n"] >= 2
        return (
            {"type": "page", "url": url, "checks": [{"name": "status", "status": "warn" if had_error else "ok"}]},
            {"url": url, "status": "warn" if had_error else "ok", "file": None, "html": None, "checks": [{"name": "status", "status": "warn" if had_error else "ok"}]},
            had_error,
        )

    monkeypatch.setattr(mod, "_process_url", _proc)
    rc = mod.run(["https://e.com/sitemap.xml", "--concurrency", "auto", "--recheck-failed", "--recheck-from", str(rp), "--no-report-ui"])
    assert rc == 0


def test_run_runtime_error_path(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(mod, "parse_sitemap_urls", lambda *_a, **_k: ["https://e.com/a"])
    monkeypatch.setattr(mod, "check_robots_txt", lambda *_a, **_k: _site_result())
    monkeypatch.setattr(mod, "check_llms_txt", lambda *_a, **_k: _llms_result())
    monkeypatch.setattr(mod, "get_page_checks", lambda **_k: [SimpleNamespace(name="status", run=lambda **_k: CheckResult(name="status", status="ok"))])

    class _BadBrowser:
        def __init__(self, **_k):
            raise RuntimeError("boom")

    monkeypatch.setattr(mod, "Browser", _BadBrowser)
    assert mod.run(["https://e.com/sitemap.xml", "--concurrency", "1", "--no-report-ui"]) == 1


def test_run_non_auto_concurrency_future_failure(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.setattr(mod, "parse_sitemap_urls", lambda *_a, **_k: ["https://e.com/a", "https://e.com/b"])
    monkeypatch.setattr(mod, "check_robots_txt", lambda *_a, **_k: _site_result())
    monkeypatch.setattr(mod, "check_llms_txt", lambda *_a, **_k: _llms_result())
    monkeypatch.setattr(mod, "get_page_checks", lambda **_k: [SimpleNamespace(name="status")])

    def _proc(url: str, **_k):
        if url.endswith("/b"):
            raise RuntimeError("boom")
        return (
            {"type": "page", "url": url, "checks": [{"name": "status", "status": "ok"}]},
            {"url": url, "status": "ok", "file": None, "html": None, "checks": [{"name": "status", "status": "ok"}]},
            False,
        )

    monkeypatch.setattr(mod, "_process_url", _proc)
    assert mod.run(["https://e.com/sitemap.xml", "--concurrency", "2", "--output-dir", str(tmp_path), "--no-report-ui"]) == 0


def test_run_auto_concurrency_full_loop(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.setattr(
        mod,
        "parse_sitemap_urls",
        lambda *_a, **_k: ["https://e.com/1", "https://e.com/2", "https://e.com/3", "https://e.com/4"],
    )
    monkeypatch.setattr(mod, "check_robots_txt", lambda *_a, **_k: _site_result())
    monkeypatch.setattr(mod, "check_llms_txt", lambda *_a, **_k: _llms_result())
    monkeypatch.setattr(mod, "get_page_checks", lambda **_k: [SimpleNamespace(name="status")])

    seen = {"n": 0}

    def _proc(url: str, **_k):
        seen["n"] += 1
        # First call no error (current grows), later calls with error (current shrinks path).
        had_error = seen["n"] >= 2
        return (
            {"type": "page", "url": url, "checks": [{"name": "status", "status": "warn" if had_error else "ok"}]},
            {"url": url, "status": "warn" if had_error else "ok", "file": None, "html": None, "checks": [{"name": "status", "status": "warn" if had_error else "ok"}]},
            had_error,
        )

    monkeypatch.setattr(mod, "_process_url", _proc)
    assert mod.run(["https://e.com/sitemap.xml", "--concurrency", "auto", "--output-dir", str(tmp_path), "--no-report-ui"]) == 0



from __future__ import annotations

import csv
import json
from pathlib import Path

import pytest
from typer.testing import CliRunner

from worai.commands import structured_data as structured_data_cmd
from worai.core.structured_data_inventory import (
    IngestLoaderUnavailableError,
    InventoryOptions,
    InventoryRow,
    build_inventory,
    build_inventory_row,
    run_inventory,
)
from worai.core.url_sources import UrlRecord
from worai.errors import UsageError


def test_build_inventory_row_extracts_faq_and_types() -> None:
    html = """
    <html><head>
      <script type="application/ld+json">
      {
        "@context": "https://schema.org",
        "@graph": [
          {"@id": "https://data.example.com/dataset/page#faq", "@type": "https://schema.org/FAQPage"},
          {"@id": "https://example.com/page", "@type": ["https://schema.org/WebPage", "Thing"]}
        ]
      }
      </script>
    </head></html>
    """

    row = build_inventory_row(
        "https://example.com/page",
        html,
        dataset_uri="https://data.example.com/dataset",
    )

    assert row.url == "https://example.com/page"
    assert row.faq_markup == "yes"
    assert row.faq_markup_from_graph == "yes"
    assert set(row.types.split(",")) == {"FAQPage", "WebPage", "Thing"}
    payload = json.loads(row.structured_data)
    assert payload["@context"] == "https://schema.org"
    assert len(payload["@graph"]) == 2


def test_build_inventory_row_ignores_invalid_jsonld_and_checks_faq_id() -> None:
    html = """
    <html><head>
      <script type="application/ld+json">{ invalid json }</script>
      <script type="application/ld+json">
      {"@type": "FAQPage", "@id": "https://elsewhere.example/faq"}
      </script>
    </head></html>
    """

    row = build_inventory_row(
        "https://example.com/page",
        html,
        dataset_uri="https://data.example.com/dataset/",
    )

    assert row.faq_markup == "yes"
    assert row.faq_markup_from_graph == "no"


def test_run_inventory_requires_exactly_one_destination() -> None:
    options = InventoryOptions(
        source="https://example.com/sitemap.xml",
        api_key="key",
    )
    with pytest.raises(UsageError):
        run_inventory(options)

    options = InventoryOptions(
        source="https://example.com/sitemap.xml",
        api_key="key",
        output="out.csv",
        destination_spreadsheet_id="spreadsheet-id",
    )
    with pytest.raises(UsageError):
        run_inventory(options)


def test_run_inventory_requires_destination_sheet_name() -> None:
    options = InventoryOptions(
        source="https://example.com/sitemap.xml",
        api_key="key",
        destination_spreadsheet_id="spreadsheet-id",
    )
    with pytest.raises(UsageError):
        run_inventory(options)


def test_run_inventory_writes_csv(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    rows = [
        InventoryRow(
            url="https://example.com/a",
            faq_markup="no",
            faq_markup_from_graph="no",
            types="WebPage",
            structured_data='{"@context":"https://schema.org","@graph":[]}',
        )
    ]

    monkeypatch.setattr("worai.core.structured_data_inventory.build_inventory", lambda _options: rows)

    output_path = tmp_path / "inventory.csv"
    run_inventory(
        InventoryOptions(
            source="https://example.com/sitemap.xml",
            api_key="key",
            output=str(output_path),
        )
    )

    with output_path.open("r", encoding="utf-8") as handle:
        reader = csv.DictReader(handle)
        records = list(reader)

    assert len(records) == 1
    assert records[0]["url"] == "https://example.com/a"
    assert records[0]["types"] == "WebPage"


def test_inventory_command_builds_options_and_emits_summary(monkeypatch: pytest.MonkeyPatch) -> None:
    captured: dict[str, object] = {}

    def _fake_run_inventory(options):
        captured["options"] = options
        return []

    monkeypatch.setattr(structured_data_cmd, "resolve_profile_api_key", lambda _ctx, explicit_profile=None: ("key", explicit_profile or "default", "worai.toml"))
    monkeypatch.setattr(structured_data_cmd, "run_inventory", _fake_run_inventory)

    runner = CliRunner()
    result = runner.invoke(
        structured_data_cmd.app,
        ["inventory", "https://example.com/sitemap.xml", "--output", "report.csv"],
    )

    assert result.exit_code == 0
    payload = json.loads(result.output)
    assert payload["total"] == 0
    options = captured["options"]
    assert isinstance(options, InventoryOptions)
    assert options.source == "https://example.com/sitemap.xml"
    assert options.output == "report.csv"
    assert options.concurrency == "auto"


def test_inventory_command_sets_destination_sheet_options(monkeypatch: pytest.MonkeyPatch) -> None:
    captured: dict[str, object] = {}

    def _fake_run_inventory(options):
        captured["options"] = options
        return []

    monkeypatch.setattr(structured_data_cmd, "resolve_profile_api_key", lambda _ctx, explicit_profile=None: ("key", explicit_profile or "default", "worai.toml"))
    monkeypatch.setattr(structured_data_cmd, "run_inventory", _fake_run_inventory)

    runner = CliRunner()
    result = runner.invoke(
        structured_data_cmd.app,
        [
            "inventory",
            "https://example.com/sitemap.xml",
            "--destination-sheet-id",
            "spreadsheet-id",
            "--destination-sheet-name",
            "Inventory",
        ],
    )

    assert result.exit_code == 0
    options = captured["options"]
    assert isinstance(options, InventoryOptions)
    assert options.destination_spreadsheet_id == "spreadsheet-id"
    assert options.destination_sheet_name == "Inventory"


def test_inventory_command_sets_concurrency(monkeypatch: pytest.MonkeyPatch) -> None:
    captured: dict[str, object] = {}

    def _fake_run_inventory(options):
        captured["options"] = options
        return []

    monkeypatch.setattr(structured_data_cmd, "resolve_profile_api_key", lambda _ctx, explicit_profile=None: ("key", explicit_profile or "default", "worai.toml"))
    monkeypatch.setattr(structured_data_cmd, "run_inventory", _fake_run_inventory)

    runner = CliRunner()
    result = runner.invoke(
        structured_data_cmd.app,
        [
            "inventory",
            "https://example.com/sitemap.xml",
            "--output",
            "report.csv",
            "--concurrency",
            "3",
        ],
    )

    assert result.exit_code == 0
    options = captured["options"]
    assert isinstance(options, InventoryOptions)
    assert options.concurrency == "3"


def test_inventory_command_sets_source_type(monkeypatch: pytest.MonkeyPatch) -> None:
    captured: dict[str, object] = {}

    def _fake_run_inventory(options):
        captured["options"] = options
        return []

    monkeypatch.setattr(structured_data_cmd, "resolve_profile_api_key", lambda _ctx, explicit_profile=None: ("key", explicit_profile or "default", "worai.toml"))
    monkeypatch.setattr(structured_data_cmd, "run_inventory", _fake_run_inventory)

    runner = CliRunner()
    result = runner.invoke(
        structured_data_cmd.app,
        [
            "inventory",
            "/tmp/debug_cloud",
            "--output",
            "report.csv",
            "--source-type",
            "debug-cloud",
        ],
    )

    assert result.exit_code == 0
    options = captured["options"]
    assert isinstance(options, InventoryOptions)
    assert options.source_type == "debug-cloud"


def test_inventory_command_sets_new_ingest_flags(monkeypatch: pytest.MonkeyPatch) -> None:
    captured: dict[str, object] = {}

    def _fake_run_inventory(options):
        captured["options"] = options
        return []

    monkeypatch.setattr(structured_data_cmd, "resolve_profile_api_key", lambda _ctx, explicit_profile=None: ("key", explicit_profile or "default", "worai.toml"))
    monkeypatch.setattr(structured_data_cmd, "run_inventory", _fake_run_inventory)

    runner = CliRunner()
    result = runner.invoke(
        structured_data_cmd.app,
        [
            "inventory",
            "https://example.com/sitemap.xml",
            "--output",
            "report.csv",
            "--ingest-source",
            "sitemap",
            "--ingest-loader",
            "web_scrape_api",
            "--no-ingest-passthrough-when-html",
        ],
    )

    assert result.exit_code == 0
    options = captured["options"]
    assert isinstance(options, InventoryOptions)
    assert options.ingest_source == "sitemap"
    assert options.ingest_loader == "web_scrape_api"
    assert options.ingest_passthrough_when_html is False


def test_build_inventory_uses_browser_with_shared_user_agent(monkeypatch: pytest.MonkeyPatch) -> None:
    captured: dict[str, object] = {}

    class _FakeResponse:
        status = 200

    class _FakePage:
        def content(self) -> str:
            return (
                "<script type='application/ld+json'>"
                "{\"@type\":\"FAQPage\",\"@id\":\"https://dataset.example/page#faq\"}"
                "</script>"
            )

        def close(self) -> None:
            return None

    class _FakeBrowser:
        def __init__(self, **kwargs):
            captured["browser_kwargs"] = kwargs

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc, tb):
            return None

        def open(self, _url: str):
            return _FakePage(), _FakeResponse(), 1.0, []

    monkeypatch.setattr(
        "worai.core.structured_data_inventory.get_dataset_uri",
        lambda _api_key, _base_url: "https://dataset.example",
    )
    monkeypatch.setattr(
        "worai.core.structured_data_inventory.resolve_url_records",
        lambda _opts: [UrlRecord(url="https://example.com/page")],
    )
    monkeypatch.setattr("worai.core.structured_data_inventory.Browser", _FakeBrowser)

    options = InventoryOptions(source="https://example.com/sitemap.xml", api_key="key")
    rows = build_inventory(options)

    assert len(rows) == 1
    assert rows[0].faq_markup == "yes"
    browser_kwargs = captured["browser_kwargs"]
    assert isinstance(browser_kwargs, dict)
    assert browser_kwargs.get("headless") is True
    assert browser_kwargs.get("user_agent")


def test_build_inventory_loader_simple_uses_requests(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(
        "worai.core.structured_data_inventory.get_dataset_uri",
        lambda _api_key, _base_url: "https://dataset.example",
    )
    monkeypatch.setattr(
        "worai.core.structured_data_inventory.resolve_url_records",
        lambda _opts: [UrlRecord(url="https://example.com/page")],
    )

    class _FakeResponse:
        status_code = 200
        text = (
            "<script type='application/ld+json'>"
            "{\"@type\":\"FAQPage\",\"@id\":\"https://dataset.example/page#faq\"}"
            "</script>"
        )

    monkeypatch.setattr("worai.core.structured_data_inventory.requests.get", lambda *args, **kwargs: _FakeResponse())

    class _FailBrowser:
        def __init__(self, **_kwargs):
            raise AssertionError("Browser should not be initialized for simple loader")

    monkeypatch.setattr("worai.core.structured_data_inventory.Browser", _FailBrowser)
    rows = build_inventory(
        InventoryOptions(
            source="https://example.com/sitemap.xml",
            api_key="key",
            ingest_loader="simple",
        )
    )
    assert len(rows) == 1
    assert rows[0].faq_markup == "yes"


def test_build_inventory_uses_embedded_html_without_browser(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(
        "worai.core.structured_data_inventory.get_dataset_uri",
        lambda _api_key, _base_url: "https://dataset.example",
    )
    monkeypatch.setattr(
        "worai.core.structured_data_inventory.resolve_url_records",
        lambda _opts: [
            UrlRecord(
                url="https://example.com/page",
                html=(
                    "<script type='application/ld+json'>"
                    "{\"@type\":\"FAQPage\",\"@id\":\"https://dataset.example/page#faq\"}"
                    "</script>"
                ),
                source_type="debug-cloud",
            )
        ],
    )

    class _FailBrowser:
        def __init__(self, **_kwargs):
            raise AssertionError("Browser should not be initialized for embedded HTML records")

    monkeypatch.setattr("worai.core.structured_data_inventory.Browser", _FailBrowser)

    options = InventoryOptions(
        source="/tmp/debug_cloud",
        source_type="debug-cloud",
        api_key="key",
    )
    rows = build_inventory(options)

    assert len(rows) == 1
    assert rows[0].faq_markup == "yes"


def test_build_inventory_playwright_unavailable_raises_typed_error(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(
        "worai.core.structured_data_inventory.get_dataset_uri",
        lambda _api_key, _base_url: "https://dataset.example",
    )
    monkeypatch.setattr(
        "worai.core.structured_data_inventory.resolve_url_records",
        lambda _opts: [UrlRecord(url="https://example.com/page")],
    )

    class _FailBrowser:
        def __init__(self, **_kwargs):
            raise RuntimeError("Playwright is not installed")

    monkeypatch.setattr("worai.core.structured_data_inventory.Browser", _FailBrowser)

    with pytest.raises(IngestLoaderUnavailableError):
        build_inventory(
            InventoryOptions(
                source="https://example.com/sitemap.xml",
                api_key="key",
                ingest_loader="playwright",
            )
        )



from __future__ import annotations

from types import SimpleNamespace

import click
import pytest
from rdflib import BNode, Graph, Literal, URIRef
from rdflib.collection import Collection

from worai.commands import validate as mod
from worai.errors import UsageError


class _R:
    def __init__(self, conforms: bool = True, warning_count: int = 0, header_graph: Graph | None = None):
        self.conforms = conforms
        self.warning_count = warning_count
        self.report_text = "RAW"
        self.report_graph = header_graph or Graph()
        self.data_graph = Graph()
        self.shape_source_map = {}


def test_validate_helpers_additional_branches(capsys: pytest.CaptureFixture[str]) -> None:
    b = BNode()
    g = Graph()
    assert mod._path_to_str(g, None) == ""
    assert mod._describe_path(g, None) == ""
    assert mod._path_to_str(g, b) == str(b)
    assert mod._shorten_term(URIRef("urn:no_delimiter")) == "urn:no_delimiter"

    shape = URIRef("urn:shape")
    alt1 = BNode()
    alt2 = BNode()
    prop1 = BNode()
    prop2 = BNode()
    or_list = BNode()
    g.add((shape, mod._SH["or"], or_list))
    Collection(g, or_list, [alt1, alt2])
    g.add((alt1, mod._SH.property, prop1))
    g.add((prop1, mod._SH.path, URIRef("http://schema.org/name")))
    g.add((alt2, mod._SH.property, prop2))
    g.add((prop2, mod._SH.path, URIRef("http://schema.org/headline")))
    hint = mod._or_shape_hint(g, shape)
    assert hint and "Provide one of" in hint

    r = _R()
    mod._emit_validation_result(r, format="raw", color=False, report_file=None)
    assert "RAW" in capsys.readouterr().out


def test_format_pretty_lines_value_shape_paths() -> None:
    rg = Graph()
    focus = URIRef("urn:focus")

    r1 = URIRef("urn:r1")
    rg.add((r1, mod._SH.resultSeverity, mod._SH.Warning))
    rg.add((r1, mod._SH.focusNode, focus))
    rg.add((r1, mod._SH.resultMessage, Literal("Value does not conform to Shape")))
    rg.add((r1, mod._SH.value, Literal("v")))

    r2 = URIRef("urn:r2")
    rg.add((r2, mod._SH.resultSeverity, mod._SH.Info))
    rg.add((r2, mod._SH.focusNode, focus))
    rg.add((r2, mod._SH.resultMessage, Literal("x must conform to one or more shapes")))

    r3 = URIRef("urn:r3")
    rg.add((r3, mod._SH.resultSeverity, URIRef("urn:unknown")))
    rg.add((r3, mod._SH.focusNode, focus))

    header, lines = mod._format_pretty_lines(rg, None, None)
    assert header.startswith("Status: WARNING") or header.startswith("Status: INFO")
    joined = "\n".join(line for _label, line in lines)
    assert "Value does not satisfy the required shape" in joined
    assert "Failed to satisfy one of the alternative constraints" in joined


def test_validate_group_resolve_command_fallback(monkeypatch: pytest.MonkeyPatch) -> None:
    group = mod._ValidateGroup(name="validate")
    calls: list[list[str]] = []

    def _fake_super(self, _ctx, args):
        calls.append(list(args))
        if args and args[0] != "legacy":
            raise click.UsageError("bad")
        return ("legacy", None, list(args))

    monkeypatch.setattr(mod.TyperGroup, "resolve_command", _fake_super)
    out = group.resolve_command(None, ["file.jsonld"])
    assert out[0] == "legacy"
    assert calls[1][0] == "legacy"

    def _always_error(self, _ctx, _args):
        raise click.UsageError("x")

    monkeypatch.setattr(mod.TyperGroup, "resolve_command", _always_error)
    with pytest.raises(click.UsageError):
        group.resolve_command(None, [])


def test_validate_jsonld_input_reraises_non_url(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(mod, "validate_file", lambda *_a, **_k: (_ for _ in ()).throw(RuntimeError("boom")))
    with pytest.raises(RuntimeError):
        mod._validate_jsonld_input("local.jsonld", None)

    with pytest.raises(UsageError):
        mod.main(SimpleNamespace(invoked_subcommand=None), list_shapes=False)


def test_validate_format_branches_and_list_shapes(monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]) -> None:
    rg = Graph()
    shape_anon = BNode()
    prop = BNode()
    rg.add((shape_anon, mod._SH.property, prop))
    rg.add((prop, mod._SH.path, URIRef("http://schema.org/name")))

    r1 = URIRef("urn:r1")
    rg.add((r1, mod._SH.resultSeverity, mod._SH.Warning))
    rg.add((r1, mod._SH.resultPath, URIRef("http://schema.org/name")))
    rg.add((r1, mod._SH.resultMessage, Literal("Less than 1 values")))
    rg.add((r1, mod._SH.sourceShape, shape_anon))

    shape_uri = URIRef("urn:/google/shape")
    rg.add((shape_uri, mod._SH.message, Literal("custom shape msg")))
    r2 = URIRef("urn:r2")
    rg.add((r2, mod._SH.resultSeverity, mod._SH.Violation))
    rg.add((r2, mod._SH.resultMessage, Literal("Value does not conform to Shape")))
    rg.add((r2, mod._SH.sourceShape, shape_uri))
    rg.add((r2, mod._SH.value, Literal("v")))

    r3 = URIRef("urn:r3")
    rg.add((r3, mod._SH.resultSeverity, mod._SH.Info))
    rg.add((r3, mod._SH.resultMessage, Literal("must conform to one or more shapes")))
    rg.add((r3, mod._SH.sourceShape, shape_uri))

    class _Result:
        conforms = True
        warning_count = 0
        report_text = "RAW"
        report_graph = rg
        data_graph = None
        shape_source_map = {}

    mod._emit_validation_result(_Result(), format="pretty", color=True, report_file=None)
    out = capsys.readouterr().out
    assert "custom shape msg" in out
    assert "[shape: (anonymous)]" in out

    monkeypatch.setitem(__import__("sys").modules, "wordlift_sdk.validation", SimpleNamespace(list_shape_names=lambda: ["s1", "s2"]))
    mod._list_shapes()
    out = capsys.readouterr().out
    assert "s1" in out and "s2" in out



"""WORAi CLI package."""

from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version

__all__ = ["__version__"]

try:
    __version__ = version("worai")
except PackageNotFoundError:
    __version__ = "0.0.0"
